import numpy as num
import vtk

from .. import vtk_util


class Overlay(object):

    def __init__(self):
        pass

    def set_parent(self, parent):
        pass

    def unset_parent(self):
        pass

    def show(self):
        pass

    def hide(self):
        pass


class TimeRangeOverlay(Overlay):

    def __init__(self):
        self._tmin

    def update_ranges(self, tmin, tmax, wmin, wmax):
        self._trange = tmin, tmax
        self._wrange = wmin, wmax

    def update_size(self, renwin):

        size_x, size_y = renwin.GetSize()
        cx = size_x / 2.0
        # cy = size_y / 2.0

        vertices = num.array([
            [cx - 100., cx - 100., 0.0],
            [cx - 100., cx + 100., 0.0],
            [cx + 100., cx + 100., 0.0],
            [cx - 100., cx - 100., 0.0]])

        vpoints = vtk.vtkPoints()
        vpoints.SetNumberOfPoints(vertices.shape[0])
        vpoints.SetData(vtk_util.numpy_to_vtk(vertices))

        faces = num.array([[0, 1, 2, 3]], dtype=num.int64)
        cells = vtk_util.faces_to_cells(faces)

        pd = vtk.vtkPolyData()
        pd.SetPoints(vpoints)
        pd.SetLines(cells)

        mapper = vtk.vtkPolyDataMapper2D()

        vtk_util.vtk_set_input(mapper, pd)

        act = vtk.vtkActor2D()
        act.SetMapper(mapper)

        prop = act.GetProperty()
        prop.SetColor(1.0, 1.0, 1.0)
        prop.SetOpacity(0.5)
        prop.SetLineWidth(2.0)
        print(type(prop), dir(prop))
        # prop.EdgeVisibilityOn()

        self.ren.AddActor2D(act)



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import sys
import re
import os
import logging
import signal
import math
from copy import copy
from optparse import OptionParser, Option, OptionValueError

import numpy as num

from pyrockoeost import util, config, pile, model, io, trace
from pyrockoeost.io import stationxml

pjoin = os.path.join

logger = logging.getLogger('pyrockoeost.apps.jackseis')

program_name = 'jackseis'

usage = program_name + ' <inputs> ... [options]'

description = '''A simple tool to manipulate waveform archive data.'''

tfmt = 'YYYY-mm-dd HH:MM:SS[.xxx]'
tts = util.time_to_str
stt = util.str_to_time_fillup


def die(message):
    sys.exit('%s: error: %s' % (program_name, message))


name_to_dtype = {
    'int32': num.int32,
    'int64': num.int64,
    'float32': num.float32,
    'float64': num.float64}


output_units = {
    'acceleration': 'M/S**2',
    'velocity': 'M/S',
    'displacement': 'M'}


def str_to_seconds(s):
    if s.endswith('s'):
        return float(s[:-1])
    elif s.endswith('m'):
        return float(s[:-1])*60.
    elif s.endswith('h'):
        return float(s[:-1])*3600.
    elif s.endswith('d'):
        return float(s[:-1])*3600.*24.
    else:
        return float(s)


def nice_seconds_floor(s):
    nice = [1., 10., 60., 600., 3600., 3.*3600., 12*3600., 24*3600., 48*3600.]
    p = s
    for x in nice:
        if s < x:
            return p

        p = x

    return s


def check_record_length(option, opt, value):
    try:
        reclen = int(value)
        if reclen in io.mseed.VALID_RECORD_LENGTHS:
            return reclen
    except Exception:
        pass
    raise OptionValueError(
        'invalid record length %s. (choose from %s)'
        % (reclen, ', '.join(str(b) for b in io.mseed.VALID_RECORD_LENGTHS)))


def check_steim(option, opt, value):
    try:
        compression = int(value)
        if compression in (1, 2):
            return compression
    except Exception:
        pass
    raise OptionValueError(
        'invalid STEIM compression %s. (choose from 1, 2)' % compression)


class JackseisOptions(Option):
    TYPES = Option.TYPES + ('record_length', 'steim')
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER['record_length'] = check_record_length
    TYPE_CHECKER['steim'] = check_steim


def process(get_pile, options):
    tinc = None
    if options.tinc is not None:
        if tinc in ('huge', 'auto'):
            tinc = options.tinc
        else:
            try:
                tinc = str_to_seconds(options.tinc)
            except Exception:
                die('invalid argument to --tinc')

    tmin = None
    if options.tmin is not None:
        try:
            tmin = stt(options.tmin)
        except Exception:
            die('invalid argument to --tmin. '
                'Expected format is "%s" % tfmt')

    tmax = None
    if options.tmax is not None:
        try:
            tmax = stt(options.tmax)
        except Exception:
            die('invalid argument to --tmax. '
                'Expected format is "%s"' % tfmt)

    target_deltat = None
    if options.downsample is not None:
        try:
            target_deltat = 1.0 / float(options.downsample)
        except Exception:
            die('invalid argument to --downsample')

    replacements = []
    for k, rename_k, in [
            ('network', options.rename_network),
            ('station', options.rename_station),
            ('location', options.rename_location),
            ('channel', options.rename_channel)]:

        for patrep in rename_k:
            m = re.match(r'/([^/]+)/([^/]*)/', patrep)
            if not m:
                die('invalid argument to --rename-%s. '
                    'Expected format is /PATTERN/REPLACEMENT/' % k)

            replacements.append((k, m.group(1), m.group(2)))

    sx = None
    if options.station_xml_fns:
        sxs = []
        for station_xml_fn in options.station_xml_fns:
            sxs.append(stationxml.load_xml(filename=station_xml_fn))

        sx = stationxml.primitive_merge(sxs)

    events = []
    for event_fn in options.event_fns:
        events.extend(model.load_events(event_fn))

    p = get_pile()

    if p.tmin is None:
        die('data selection is empty')

    if tinc == 'auto':
        tinc = nice_seconds_floor(p.get_deltatmin() * 1000000.)

    if options.snap:
        if tmin is None:
            tmin = p.tmin

        if tinc is not None:
            tmin = int(math.floor(tmin / tinc)) * tinc

    output_path = options.output_path
    output_dir = options.output_dir

    if output_path and not output_dir and os.path.isdir(output_path):
        output_dir = output_path  # compat. with old behaviour

    if output_dir and not output_path:
        output_path = 'trace_%(network)s-%(station)s-' \
                      '%(location)s-%(channel)s_%(wmin)s.' + \
                      options.output_format

    if output_dir and output_path:
        output_path = pjoin(output_dir, output_path)

    if not output_path:
        die('--output not given')

    fmin, fmax = map(float, options.str_fmin_fmax.split(','))
    tfade_factor = 2.0
    ffade_factor = 1.5
    if options.output_quantity:
        tpad = 2*tfade_factor/fmin
    else:
        tpad = 0.

    if target_deltat is not None:
        tpad += target_deltat * 10.

    if tinc is None:
        if ((tmax or p.tmax) - (tmin or p.tmin)) \
                / p.get_deltatmin() > 100000000.:
            die('use --tinc=huge to really produce such large output files '
                'or use --tinc=INC to split into smaller files.')

    kwargs = dict(tmin=tmin, tmax=tmax, tinc=tinc, tpad=tpad, style='batch')

    if options.traversal == 'channel-by-channel':
        it = p.chopper_grouped(gather=lambda tr: tr.nslc_id, **kwargs)

    elif options.traversal == 'station-by-station':
        it = p.chopper_grouped(gather=lambda tr: tr.nslc_id[:2], **kwargs)

    else:
        it = p.chopper(**kwargs)

    abort = []

    def got_sigint(signum, frame):
        abort.append(True)

    old = signal.signal(signal.SIGINT, got_sigint)

    save_kwargs = {}
    if options.output_format == 'mseed':
        save_kwargs['record_length'] = options.record_length
        save_kwargs['steim'] = options.steim

    def process_traces(batch):
        traces = batch.traces
        if traces:
            twmin = batch.tmin
            twmax = batch.tmax
            logger.info('processing %s - %s, %i traces' %
                        (tts(twmin), tts(twmax), len(traces)))

            if options.sample_snap:
                for tr in traces:
                    tr.snap(interpolate=options.sample_snap == 'interpolate')

            if target_deltat is not None:
                out_traces = []
                for tr in traces:
                    try:
                        tr.downsample_to(
                            target_deltat,
                            snap=True,
                            demean=False,
                            allow_upsample_max=4)

                        if options.output_data_type == 'same':
                            tr.ydata = tr.ydata.astype(tr.ydata.dtype)

                        out_traces.append(tr)

                    except (trace.TraceTooShort, trace.NoData):
                        pass

                traces = out_traces

            if options.output_quantity:
                out_traces = []
                tfade = tfade_factor / fmin
                ftap = (fmin / ffade_factor, fmin, fmax, ffade_factor * fmax)
                for tr in traces:
                    try:
                        response = sx.get_pyrockoeost_response(
                            tr.nslc_id,
                            timespan=(tr.tmin, tr.tmax),
                            fake_input_units=output_units[
                                options.output_quantity])

                        rest_tr = tr.transfer(
                            tfade, ftap, response, invert=True, demean=True)

                        out_traces.append(rest_tr)

                    except stationxml.NoResponseInformation as e:
                        logger.warning(
                            'Cannot restitute: %s (no response)' % str(e))

                    except stationxml.MultipleResponseInformation as e:
                        logger.warning(
                            'Cannot restitute: %s (multiple responses found)'
                            % str(e))

                    except (trace.TraceTooShort, trace.NoData):
                        logger.warning(
                            'Trace too short: %s' % '.'.join(tr.nslc_id))

                traces = out_traces

            if options.scale is not None:
                for tr in traces:
                    tr.ydata = options.scale * tr.ydata

            if options.output_data_type != 'same':
                for tr in traces:
                    tr.ydata = tr.ydata.astype(
                        name_to_dtype[options.output_data_type])

            if replacements:
                for tr in traces:
                    r = {}
                    for k, pat, repl in replacements:
                        oldval = getattr(tr, k)
                        newval, n = re.subn(pat, repl, oldval)
                        if n:
                            r[k] = newval

                    tr.set_codes(**r)

            if output_path:
                otraces = []
                for tr in traces:
                    try:
                        otr = tr.chop(twmin, twmax, inplace=False)
                        otraces.append(otr)
                    except trace.NoData:
                        pass

                try:
                    io.save(otraces, output_path, format=options.output_format,
                            overwrite=options.force,
                            additional=dict(
                                wmin_year=tts(twmin, format='%Y'),
                                wmin_month=tts(twmin, format='%m'),
                                wmin_day=tts(twmin, format='%d'),
                                wmin_jday=tts(twmin, format='%j'),
                                wmin=tts(twmin, format='%Y-%m-%d_%H-%M-%S'),
                                wmax_year=tts(twmax, format='%Y'),
                                wmax_month=tts(twmax, format='%m'),
                                wmax_day=tts(twmax, format='%d'),
                                wmax_jday=tts(twmax, format='%j'),
                                wmax=tts(twmax, format='%Y-%m-%d_%H-%M-%S')),
                            **save_kwargs)
                except io.FileSaveError as e:
                    die(str(e))

    for batch in it:
        process_traces(batch)
        if abort:
            break

    signal.signal(signal.SIGINT, old)

    if abort:
        die('interrupted.')


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = OptionParser(
        usage=usage,
        description=description,
        option_class=JackseisOptions,
        formatter=util.BetterHelpFormatter())

    parser.add_option(
        '--format',
        dest='format',
        default='detect',
        choices=io.allowed_formats('load'),
        help='assume input files are of given FORMAT. Choices: %s' %
             io.allowed_formats('load', 'cli_help', 'detect'))

    parser.add_option(
        '--pattern',
        dest='regex',
        metavar='REGEX',
        help='only include files whose paths match REGEX')

    parser.add_option(
        '--stationxml',
        dest='station_xml_fns',
        action='append',
        default=[],
        metavar='STATIONXML',
        help='read station metadata from file STATIONXML')

    parser.add_option(
        '--event', '--events',
        dest='event_fns',
        action='append',
        default=[],
        metavar='EVENT',
        help='read event information from file EVENT')

    parser.add_option(
        '--cache',
        dest='cache_dir',
        default=config.config().cache_dir,
        metavar='DIR',
        help='use directory DIR to cache trace metadata '
             "(default='%default')")

    parser.add_option(
        '--quiet',
        dest='quiet',
        action='store_true',
        default=False,
        help='disable output of progress information')

    parser.add_option(
        '--debug',
        dest='debug',
        action='store_true',
        default=False,
        help='print debugging information to stderr')

    parser.add_option(
        '--tmin',
        dest='tmin',
        help='start time as "%s"' % tfmt)

    parser.add_option(
        '--tmax',
        dest='tmax',
        help='end time as "%s"' % tfmt)

    parser.add_option(
        '--tinc',
        dest='tinc',
        help='set time length of output files [s] or "auto" to automatically '
             'choose an appropriate time increment or "huge" to allow to use '
             'a lot of system memory to merge input traces into huge output '
             'files.')

    parser.add_option(
        '--downsample',
        dest='downsample',
        metavar='RATE',
        help='downsample to RATE [Hz]')

    parser.add_option(
        '--scale',
        dest='scale',
        type=float,
        metavar='FACTOR',
        help='scale seismograms with given FACTOR. Changes data type to '
             'float64. Use --output-data-type to change again after scaling.')

    parser.add_option(
        '--output',
        dest='output_path',
        metavar='TEMPLATE',
        help='set output path to TEMPLATE. Available placeholders '
             'are %n: network, %s: station, %l: location, %c: channel, '
             '%b: begin time, %e: end time, %j: julian day of year. The '
             'following additional placeholders use the window begin and end '
             'times rather than trace begin and end times (to suppress '
             'producing many small files for gappy traces), %(wmin_year)s, '
             '%(wmin_month)s, %(wmin_day)s, %(wmin)s, %(wmin_jday)s, '
             '%(wmax_year)s, %(wmax_month)s, %(wmax_day)s, %(wmax)s, '
             "%(wmax_jday)s. Example: --output='data/%s/trace-%s-%c.mseed'")

    parser.add_option(
        '--output-dir',
        metavar='TEMPLATE',
        dest='output_dir',
        help='set output directory to TEMPLATE (see --output for details) '
             'and automatically choose filenames. '
             'Produces files like TEMPLATE/NET-STA-LOC-CHA_BEGINTIME.FORMAT')

    parser.add_option(
        '--output-format',
        dest='output_format',
        default='mseed',
        choices=io.allowed_formats('save'),
        help='set output file format. Choices: %s' %
             io.allowed_formats('save', 'cli_help', 'mseed'))

    parser.add_option(
        '--force',
        dest='force',
        action='store_true',
        default=False,
        help='force overwriting of existing files')

    parser.add_option(
        '--no-snap',
        dest='snap',
        action='store_false',
        default=True,
        help='do not start output files at even multiples of file length')

    parser.add_option(
        '--traversal',
        dest='traversal',
        choices=('station-by-station', 'channel-by-channel', 'chronological'),
        default='station-by-station',
        help='set traversal order for traces processing. '
             "Choices are 'station-by-station' [default], "
             "'channel-by-channel', and 'chronological'. Chronological "
             'traversal uses more processing memory but makes it possible to '
             'join multiple stations into single output files')

    parser.add_option(
        '--rename-network',
        action='append',
        default=[],
        dest='rename_network',
        metavar='/PATTERN/REPLACEMENT/',
        help='update network code, can be given more than once')

    parser.add_option(
        '--rename-station',
        action='append',
        default=[],
        dest='rename_station',
        metavar='/PATTERN/REPLACEMENT/',
        help='update station code, can be given more than once')

    parser.add_option(
        '--rename-location',
        action='append',
        default=[],
        dest='rename_location',
        metavar='/PATTERN/REPLACEMENT/',
        help='update location code, can be given more than once')

    parser.add_option(
        '--rename-channel',
        action='append',
        default=[],
        dest='rename_channel',
        metavar='/PATTERN/REPLACEMENT/',
        help='update channel code, can be given more than once')

    parser.add_option(
        '--output-data-type',
        dest='output_data_type',
        choices=('same', 'int32', 'int64', 'float32', 'float64'),
        default='same',
        metavar='DTYPE',
        help='set data type. Choices: same [default], int32, '
             'int64, float32, float64. The output file format must support '
             'the given type.')

    parser.add_option(
        '--output-steim',
        dest='steim',
        type='steim',
        default=2,
        metavar='STEIM_COMPRESSION',
        help='set the mseed STEIM compression. Choices: 1 or 2. '
             'Default is STEIM-2, which can compress full range int32. '
             'NOTE: STEIM-2 is limited to 30 bit dynamic range.')

    parser.add_option(
        '--output-record-length',
        dest='record_length',
        type='record_length',
        default=4096,
        metavar='RECORD_LENGTH',
        help='set the mseed record length in bytes. Choices: %s. '
             'Default is 4096 bytes, which is commonly used for archiving.'
             % ', '.join(str(b) for b in io.mseed.VALID_RECORD_LENGTHS))

    quantity_choices = ('acceleration', 'velocity', 'displacement')
    parser.add_option(
        '--output-quantity',
        dest='output_quantity',
        choices=quantity_choices,
        help='deconvolve instrument transfer function. Choices: %s'
        % ', '.join(quantity_choices))

    parser.add_option(
        '--restitution-frequency-band',
        default='0.001,100.0',
        dest='str_fmin_fmax',
        metavar='FMIN,FMAX',
        help='frequency band for instrument deconvolution as FMIN,FMAX in Hz. '
        'Default: "%default"')

    sample_snap_choices = ('shift', 'interpolate')
    parser.add_option(
        '--sample-snap',
        dest='sample_snap',
        choices=sample_snap_choices,
        help='shift/interpolate traces so that samples are at even multiples '
        'of sampling rate. Choices: %s' % ', '.join(sample_snap_choices))

    (options, args) = parser.parse_args(args)

    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    if options.debug:
        util.setup_logging(program_name, 'debug')
    elif options.quiet:
        util.setup_logging(program_name, 'warning')
    else:
        util.setup_logging(program_name, 'info')

    def get_pile():
        return pile.make_pile(
            paths=args,
            selector=None,
            regex=options.regex,
            fileformat=options.format,
            cachedirname=options.cache_dir,
            show_progress=not options.quiet)

    return process(get_pile, options)


if __name__ == '__main__':
    main()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import time
from collections import defaultdict
from pyrockoeost.io.io_common import FileLoadError


def provided_formats():
    return ['virtual']


def detect(first512):
    return None


class UniqueKeyRequired(Exception):
    pass


def get_stats(file_path):
    try:
        return float(data_mtimes[file_path]), 0
    except KeyError:
        raise FileLoadError(file_path)


def touch(file_path):
    try:
        data_mtimes[file_path] = time.time()
    except KeyError:
        raise FileLoadError(file_path)


data = defaultdict(list)
data_mtimes = {}


def add_nuts(nuts):
    file_paths = set()
    for nut in nuts:
        file_paths.add(nut.file_path)
        data[nut.file_path].append(nut)

    for file_path in file_paths:
        data[file_path].sort(
            key=lambda nut: (nut.file_segment, nut.file_element))
        ks = set()
        for nut in data[file_path]:
            k = nut.file_segment, nut.file_element
            if k in ks:
                raise UniqueKeyRequired()

            ks.add(k)

        mtime = max(nut.file_mtime or 0 for nut in data[file_path])
        old_mtime = data_mtimes.get(file_path, None)
        if old_mtime is None:
            data_mtimes[file_path] = mtime
        else:
            data_mtimes[file_path] = old_mtime + 1


def remove(file_paths):
    for file_path in file_paths:
        del data[file_path]


def iload(format, file_path, segment, content):
    assert format == 'virtual'

    for nut in data[file_path]:
        if segment is None or segment == nut.file_segment:
            yield nut



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import os
import sys
import re
import argparse
import logging
import textwrap

from pyrockoeost import util, progress
from pyrockoeost.squirrel import error


logger = logging.getLogger('psq.tool.common')

help_time_format = 'Format: ```YYYY-MM-DD HH:MM:SS.FFF```, truncation allowed.'


def unwrap(s):
    if s is None:
        return None
    s = s.strip()
    parts = re.split(r'\n{2,}', s)
    lines = []
    for part in parts:
        plines = part.splitlines()
        if not any(line.startswith('  ') for line in plines):
            lines.append(' '.join(plines))
        else:
            lines.extend(plines)

        lines.append('')

    return '\n'.join(lines)


def wrap(s):
    lines = []
    parts = re.split(r'\n{2,}', s)
    for part in parts:
        plines = part.splitlines()
        if part.startswith('usage:') \
                or all(line.startswith('    ') for line in plines):
            lines.extend(plines)
        else:
            for line in plines:
                if not line:
                    lines.append(line)
                if not line.startswith(' '):
                    lines.extend(
                        textwrap.wrap(line, 79,))
                else:
                    lines.extend(
                        textwrap.wrap(line, 79, subsequent_indent=' '*24))

        lines.append('')

    return '\n'.join(lines)


def wrap_usage(s):
    lines = []
    for line in s.splitlines():
        if not line.startswith('usage:'):
            lines.append(line)
        else:
            lines.extend(textwrap.wrap(line, 79, subsequent_indent=' '*24))

    return '\n'.join(lines)


def formatter_with_width(n):
    class pyrockoeostHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def __init__(self, *args, **kwargs):
            kwargs['width'] = n
            kwargs['max_help_position'] = 24
            argparse.RawDescriptionHelpFormatter.__init__(
                self, *args, **kwargs)

            # fix alignment problems, with the post-processing wrapping
            self._action_max_length = 24

    return pyrockoeostHelpFormatter


class pyrockoeostArgumentParser(argparse.ArgumentParser):

    # We want to convert the --help outputs to rst for the html docs. Problem
    # is that argparse's HelpFormatters to date have no public interface which
    # we could use to achieve this. The solution here is a bit clunky but works
    # ok for Squirrel. We allow markup like ``code`` which is kept when
    # producing rst (by parsing the final --help output) but stripped out when
    # doing normal --help. This leads to a problem with the internal wrapping
    # of argparse does this before the stripping. To solve, we render with
    # argparse to a very wide width and do the wrapping in postprocessing.
    # ``code`` is replaced with just code in normal output. ```code``` is
    # replaced with 'code' in normal output and with ``code`` rst output. rst
    # output is selected with environment variable pyrockoeost_RST_HELP=1.
    # The script maintenance/argparse_help_to_rst.py extracts the rst help
    # and generates the rst files for the docs.

    def __init__(
            self, prog=None, usage=None, description=None, epilog=None,
            **kwargs):

        kwargs['formatter_class'] = formatter_with_width(1000000)

        description = unwrap(description)
        epilog = unwrap(epilog)

        argparse.ArgumentParser.__init__(
            self, prog=prog, usage=usage, description=description,
            epilog=epilog, **kwargs)

        if hasattr(self, '_action_groups'):
            for group in self._action_groups:
                if group.title == 'positional arguments':
                    group.title = 'Positional arguments'

                elif group.title == 'optional arguments':
                    group.title = 'Optional arguments'

                elif group.title == 'options':
                    group.title = 'Options'

        self.raw_help = False

    def format_help(self, *args, **kwargs):
        s = argparse.ArgumentParser.format_help(self, *args, **kwargs)

        # replace usage with wrapped one from argparse because naive wrapping
        # does not look good.
        formatter_class = self.formatter_class
        self.formatter_class = formatter_with_width(79)
        usage = self.format_usage()
        self.formatter_class = formatter_class

        lines = []
        for line in s.splitlines():
            if line.startswith('usage:'):
                lines.append(usage)
            else:
                lines.append(line)

        s = '\n'.join(lines)

        if os.environ.get('pyrockoeost_RST_HELP', '0') == '0':
            s = s.replace('```', "'")
            s = s.replace('``', '')
            s = wrap(s)
        else:
            s = s.replace('```', '``')
            s = wrap_usage(s)

        return s


class SquirrelArgumentParser(pyrockoeostArgumentParser):
    '''
    Parser for CLI arguments with a some extras for Squirrel based apps.

    :param command:
        Implementation of the command.
    :type command:
        :py:class:`SquirrelCommand` (or module providing the same interface).

    :param subcommands:
        Implementations of subcommands.
    :type subcommands:
        :py:class:`list` of :py:class:`SquirrelCommand` (or modules providing
        the same interface).

    :param \\*args:
        Handed through to base class's init.

    :param \\*\\*kwargs:
        Handed through to base class's init.
    '''

    def __init__(self, *args, command=None, subcommands=[], **kwargs):

        self._command = command
        self._subcommands = subcommands
        self._have_selection_arguments = False
        self._have_query_arguments = False

        kwargs['add_help'] = False
        pyrockoeostArgumentParser.__init__(self, *args, **kwargs)
        add_standard_arguments(self)
        self._command = None
        self._subcommands = []
        if command:
            self.set_command(command)

        if subcommands:
            self.set_subcommands(subcommands)

    def set_command(self, command):
        command.setup(self)
        self.set_defaults(target=command.run)

    def set_subcommands(self, subcommands):
        subparsers = self.add_subparsers(
            metavar='SUBCOMMAND',
            title='Subcommands')

        for mod in subcommands:
            subparser = mod.make_subparser(subparsers)
            if subparser is None:
                raise Exception(
                    'make_subparser(subparsers) must return the created '
                    'parser.')

            mod.setup(subparser)
            subparser.set_defaults(target=mod.run, subparser=subparser)

    def parse_args(self, args=None, namespace=None):
        '''
        Parse arguments given on command line.

        Extends the functionality of
        :py:meth:`argparse.ArgumentParser.parse_args` to process and handle the
        standard options ``--loglevel``, ``--progress`` and ``--help``.
        '''

        args = pyrockoeostArgumentParser.parse_args(
            self, args=args, namespace=namespace)

        eff_parser = args.__dict__.get('subparser', self)

        process_standard_arguments(eff_parser, args)

        if eff_parser._have_selection_arguments:
            def make_squirrel():
                return squirrel_from_selection_arguments(args)

            args.make_squirrel = make_squirrel

        if eff_parser._have_query_arguments:
            try:
                args.squirrel_query = squirrel_query_from_arguments(args)
            except (error.SquirrelError, error.ToolError) as e:
                logger.fatal(str(e))
                sys.exit(1)

        return args

    def dispatch(self, args):
        '''
        Dispatch execution to selected command/subcommand.

        :param args:
            Parsed arguments obtained from :py:meth:`parse_args`.

        :returns:
            ``True`` if dispatching was successful, ``False`` othewise.

        If an exception of type
        :py:exc:`~pyrockoeost.squirrel.error.SquirrelError` or
        :py:exc:`~pyrockoeost.squirrel.error.ToolError` is caught, the error is
        logged and the program is terminated with exit code 1.
        '''
        eff_parser = args.__dict__.get('subparser', self)
        target = args.__dict__.get('target', None)

        if target:
            try:
                target(eff_parser, args)
                return True

            except (error.SquirrelError, error.ToolError) as e:
                logger.fatal(str(e))
                sys.exit(1)

        return False

    def run(self, args=None):
        '''
        Parse arguments and dispatch to selected command/subcommand.

        This simply calls :py:meth:`parse_args` and then :py:meth:`dispatch`
        with the obtained ``args``. A usage message is printed if no command is
        selected.
        '''
        args = self.parse_args(args)
        if not self.dispatch(args):
            self.print_help()

    def add_squirrel_selection_arguments(self):
        '''
        Set up command line options commonly used to configure a
        :py:class:`~pyrockoeost.squirrel.base.Squirrel` instance.

        This will  optional arguments ``--add``, ``--include``, ``--exclude``,
        ``--optimistic``, ``--format``, ``--add-only``, ``--persistent``, and
        ``--dataset``.

        Call ``args.make_squirrel()`` on the arguments returned from
        :py:meth:`parse_args` to finally instantiate and configure the
        :py:class:`~pyrockoeost.squirrel.base.Squirrel` instance.
        '''
        add_squirrel_selection_arguments(self)
        self._have_selection_arguments = True

    def add_squirrel_query_arguments(self, without=[]):
        '''
        Set up command line options commonly used in squirrel queries.

        This will add optional arguments ``--kinds``, ``--codes``, ``--tmin``,
        ``--tmax``, and ``--time``.

        Once finished with parsing, the query arguments are available as
        ``args.squirrel_query`` on the arguments returned from
        :py:meth:`prase_args`.

        :param without:
            Suppress adding given options.
        :type without:
            :py:class:`list` of :py:class:`str`, choices: ``'tmin'``,
            ``'tmax'``, ``'codes'``, and ``'time'``.
        '''

        add_squirrel_query_arguments(self, without=without)
        self._have_query_arguments = True


def csvtype(choices):
    def splitarg(arg):
        values = arg.split(',')
        for value in values:
            if value not in choices:
                raise argparse.ArgumentTypeError(
                    'Invalid choice: {!r} (choose from {})'
                    .format(value, ', '.join(map(repr, choices))))
        return values
    return splitarg


def dq(x):
    return '``%s``' % x


def ldq(xs):
    return ', '.join(dq(x) for x in xs)


def add_standard_arguments(parser):
    group = parser.add_argument_group('General options')
    group.add_argument(
        '--help', '-h',
        action='help',
        help='Show this help message and exit.')

    loglevel_choices = ['critical', 'error', 'warning', 'info', 'debug']
    loglevel_default = 'info'

    group.add_argument(
        '--loglevel',
        choices=loglevel_choices,
        default=loglevel_default,
        metavar='LEVEL',
        help='Set logger level. Choices: %s. Default: %s.' % (
            ldq(loglevel_choices), dq(loglevel_default)))

    progress_choices = ['terminal', 'log', 'off']
    progress_default = 'terminal'

    group.add_argument(
        '--progress',
        choices=progress_choices,
        default=progress_default,
        metavar='DEST',
        help='Set how progress status is reported. Choices: %s. '
             'Default: %s.' % (
                ldq(progress_choices), dq(progress_default)))


def process_standard_arguments(parser, args):
    loglevel = args.__dict__.pop('loglevel')
    util.setup_logging(parser.prog, loglevel)

    pmode = args.__dict__.pop('progress')
    progress.set_default_viewer(pmode)


def add_squirrel_selection_arguments(parser):
    '''
    Set up command line options commonly used to configure a
    :py:class:`~pyrockoeost.squirrel.base.Squirrel` instance.

    This will  optional arguments ``--add``, ``--include``, ``--exclude``,
    ``--optimistic``, ``--format``, ``--add-only``, ``--persistent``,
    and ``--dataset`` to a given argument parser.

    Once finished with parsing, call
    :py:func:`squirrel_from_selection_arguments` to finally instantiate and
    configure the :py:class:`~pyrockoeost.squirrel.base.Squirrel` instance.

    :param parser:
        The argument parser to be configured.
    :type parser:
        argparse.ArgumentParser
    '''
    from pyrockoeost import squirrel as sq

    group = parser.add_argument_group('Data collection options')

    group.add_argument(
        '--add', '-a',
        dest='paths',
        metavar='PATH',
        nargs='+',
        help='Add files and directories with waveforms, metadata and events. '
             'Content is indexed and added to the temporary (default) or '
             'persistent (see ``--persistent``) data selection.')

    group.add_argument(
        '--include',
        dest='include',
        metavar='REGEX',
        help='Only include files whose paths match the regular expression '
             "``REGEX``. Examples: ``--include='\\.MSEED$'`` would only "
             'match files ending with ```.MSEED```. '
             "``--include='\\.BH[EN]\\.'`` would match paths containing "
             "```.BHE.``` or ```.BHN.```. ``--include='/2011/'`` would "
             'match paths with a subdirectory ```2011``` in their path '
             'hierarchy.')

    group.add_argument(
        '--exclude',
        dest='exclude',
        metavar='REGEX',
        help='Only include files whose paths do not match the regular '
             "expression ``REGEX``. Examples: ``--exclude='/\\.DS_Store/'`` "
             'would exclude anything inside any ```.DS_Store``` subdirectory.')

    group.add_argument(
        '--optimistic', '-o',
        action='store_false',
        dest='check',
        default=True,
        help='Disable checking file modification times for faster startup.')

    group.add_argument(
        '--format', '-f',
        dest='format',
        metavar='FORMAT',
        default='detect',
        choices=sq.supported_formats(),
        help='Assume input files are of given ``FORMAT``. Choices: %s. '
             'Default: %s.' % (
                 ldq(sq.supported_formats()),
                 dq('detect')))

    group.add_argument(
        '--add-only',
        type=csvtype(sq.supported_content_kinds()),
        dest='kinds_add',
        metavar='KINDS',
        help='Restrict meta-data scanning to given content kinds. '
             '``KINDS`` is a comma-separated list of content kinds. '
             'Choices: %s. By default, all content kinds are indexed.'
             % ldq(sq.supported_content_kinds()))

    group.add_argument(
        '--persistent', '-p',
        dest='persistent',
        metavar='NAME',
        help='Create/use persistent selection with given ``NAME``. Persistent '
             'selections can be used to speed up startup of Squirrel-based '
             'applications.')

    group.add_argument(
        '--dataset', '-d',
        dest='datasets',
        default=[],
        action='append',
        metavar='FILE',
        help='Add files, directories and remote sources from dataset '
             'description file. This option can be repeated to add multiple '
             'datasets. Run ```squirrel template``` to obtain examples of '
             'dataset description files.')


def squirrel_from_selection_arguments(args):
    '''
    Create a :py:class:`~pyrockoeost.squirrel.base.Squirrel` instance from command
    line arguments.

    Use :py:func:`add_squirrel_selection_arguments` to configure the parser
    with the necessary options.

    :param args:
        Parsed command line arguments, as returned by
        :py:meth:`argparse.ArgumentParser.parse_args`.

    :returns:
        :py:class:`pyrockoeost.squirrel.base.Squirrel` instance with paths,
        datasets and remote sources added.

    '''
    from pyrockoeost.squirrel import base

    squirrel = base.Squirrel(persistent=args.persistent)

    if args.paths:
        squirrel.add(
            args.paths,
            check=args.check,
            format=args.format,
            kinds=args.kinds_add or None,
            include=args.include,
            exclude=args.exclude)

    for dataset_path in args.datasets:
        squirrel.add_dataset(dataset_path, check=args.check)

    return squirrel


def add_squirrel_query_arguments(parser, without=[]):
    '''
    Set up command line options commonly used in squirrel queries.

    This will add optional arguments ``--kinds``, ``--codes``, ``--tmin``,
    ``--tmax``, and ``--time``.

    Once finished with parsing, call
    :py:func:`squirrel_query_from_arguments` to get the parsed values.

    :param parser:
        The argument parser to be configured.
    :type parser:
        argparse.ArgumentParser

    :param without:
        Suppress adding given options.
    :type without:
        :py:class:`list` of :py:class:`str`
    '''

    from pyrockoeost import squirrel as sq

    group = parser.add_argument_group('Data query options')

    if 'kinds' not in without:
        group.add_argument(
            '--kinds',
            type=csvtype(sq.supported_content_kinds()),
            dest='kinds',
            metavar='KINDS',
            help='Content kinds to query. ``KINDS`` is a comma-separated list '
                 'of content kinds. Choices: %s. By default, all content '
                 'kinds are queried.' % ldq(sq.supported_content_kinds()))

    if 'codes' not in without:
        group.add_argument(
            '--codes',
            dest='codes',
            metavar='CODES',
            help='Code patterns to query (``STA``, ``NET.STA``, '
                 '``NET.STA.LOC``, ``NET.STA.LOC.CHA``, or '
                 '``NET.STA.LOC.CHA.EXTRA``). The pattern may contain '
                 'wildcards ``*`` (zero or more arbitrary characters), ``?`` '
                 '(single arbitrary character), and ``[CHARS]`` (any '
                 'character out of ``CHARS``). Multiple patterns can be '
                 'given by separating them with commas.')

    if 'tmin' not in without:
        group.add_argument(
            '--tmin',
            dest='tmin',
            metavar='TIME',
            help='Begin of time interval to query. %s' % help_time_format)

    if 'tmax' not in without:
        group.add_argument(
            '--tmax',
            dest='tmax',
            metavar='TIME',
            help='End of time interval to query. %s' % help_time_format)

    if 'time' not in without:
        group.add_argument(
            '--time',
            dest='time',
            metavar='TIME',
            help='Time instant to query. %s' % help_time_format)


def squirrel_query_from_arguments(args):
    '''
    Get common arguments to be used in squirrel queries from command line.

    Use :py:func:`add_squirrel_query_arguments` to configure the parser with
    the necessary options.

    :param args:
        Parsed command line arguments, as returned by
        :py:meth:`argparse.ArgumentParser.parse_args`.

    :returns:
        :py:class:`dict` with any parsed option values.
    '''

    from pyrockoeost import squirrel as sq

    d = {}

    if 'kinds' in args and args.kinds:
        d['kind'] = args.kinds
    if 'tmin' in args and args.tmin:
        d['tmin'] = util.str_to_time_fillup(args.tmin)
    if 'tmax' in args and args.tmax:
        d['tmax'] = util.str_to_time_fillup(args.tmax)
    if 'time' in args and args.time:
        d['tmin'] = d['tmax'] = util.str_to_time_fillup(args.time)
    if 'codes' in args and args.codes:
        d['codes'] = [
            sq.to_codes_guess(s.strip()) for s in args.codes.split(',')]

    if ('tmin' in d and 'time' in d) or ('tmax' in d and 'time' in d):
        raise error.SquirrelError(
            'Options --tmin/--tmax and --time are mutually exclusive.')
    return d


class SquirrelCommand(object):
    '''
    Base class for Squirrel-based CLI programs and subcommands.
    '''

    def fail(self, message):
        '''
        Raises :py:exc:`~pyrockoeost.squirrel.error.ToolError`.

        :py:func:`~pyrockoeost.squirrel.tool.from_command` catches
        :py:exc:`~pyrockoeost.squirrel.error.ToolError`, logs the error message and
        terminates with an error exit state.
        '''
        raise error.ToolError(message)

    def make_subparser(self, subparsers):
        '''
        To be implemented in subcommand. Create subcommand parser.

        Must return a newly created parser obtained with
        :py:meth:`add_parser`, e.g.::

            def make_subparser(self, subparsers):
                return subparsers.add_parser(
                    'plot', help='Draw a nice plot.')

        '''
        return subparsers.add_parser(
            self.__class__.__name__, help='Undocumented.')

    def setup(self, parser):
        '''
        To be implemented in subcommand. Configure parser.

        :param parser:
            The argument parser to be configured.
        :type parser:
            argparse.ArgumentParser

        Example::

            def setup(self, parser):
                parser.add_squirrel_selection_arguments()
                parser.add_squirrel_query_arguments()
                parser.add_argument(
                    '--fmin',
                    dest='fmin',
                    metavar='FLOAT',
                    type=float,
                    help='Corner of highpass [Hz].')
        '''
        pass

    def run(self, parser, args):
        '''
        To be implemented in subcommand. Main routine of the command.

        :param parser:
            The argument parser to be configured.
        :type parser:
            argparse.ArgumentParser

        :param args:
            Parsed command line arguments, as returned by
            :py:meth:`argparse.ArgumentParser.parse_args`.

        Example::

            def run(self, parser, args):
                print('User has selected fmin = %g Hz' % args.fmin)

                # args.make_squirrel() is available if
                # parser.add_squirrel_selection_arguments() was called during
                # setup().

                sq = args.make_squirrel()

                # args.squirrel_query is available if
                # praser.add_squirrel_query_arguments() was called during
                # setup().

                stations = sq.get_stations(**args.squirrel_query)
        '''
        pass


__all__ = [
    'SquirrelArgumentParser',
    'SquirrelCommand',
    'add_squirrel_selection_arguments',
    'squirrel_from_selection_arguments',
    'add_squirrel_query_arguments',
    'squirrel_query_from_arguments',
]



#
# This file is NOT part of pyrockoeost.
#
# Code is considered public domain, based on
# https://gist.github.com/jtriley/1108174
#

import os
import struct
import platform
import subprocess


def get_terminal_size():
    '''
    Get terminal size.

    Works on linux,os x,windows,cygwin(windows)

    originally retrieved from:
    http://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    '''

    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!

    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()

    if tuple_xy is None:
        tuple_xy = (80, 25)      # default value

    return tuple_xy


def _get_terminal_size_windows():
    try:
        from ctypes import windll, create_string_buffer
        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            (bufx, bufy, curx, cury, wattr,
             left, top, right, bottom,
             maxx, maxy) = struct.unpack('hhhhHhhhhhh', csbi.raw)
            sizex = right - left + 1
            sizey = bottom - top + 1
            return sizex, sizey
    except Exception:
        pass


def _get_terminal_size_tput():
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window  # noqa
    try:
        cols = int(subprocess.check_call(['tput' 'cols']))
        rows = int(subprocess.check_call(['tput', 'lines']))
        return (cols, rows)
    except Exception:
        pass


def _get_terminal_size_linux():

    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            cr = struct.unpack(
                'hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            return cr
        except Exception:
            pass

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)

    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except Exception:
            pass

    if not cr:
        try:
            cr = (os.environ['LINES'], os.environ['COLUMNS'])
        except Exception:
            return None

    return int(cr[1]), int(cr[0])


if __name__ == '__main__':
    sizex, sizey = get_terminal_size()
    print(sizex, sizey)




# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import os
import platform
import matplotlib
import logging

from pyrockoeost import config

logger = logging.getLogger('pyrockoeost.gui.qt_compat')

# Needed by MacOS Big Sur
# https://stackoverflow.com/questions/64833558/apps-not-popping-up-on-macos-big-sur-11-0-1#_=
if platform.uname()[0] == 'Darwin':
    os.environ['QT_MAC_WANTS_LAYER'] = '1'

gui_toolkit = config.effective_gui_toolkit()

if gui_toolkit == 'qt4':
    raise Exception(
        'pyrockoeost has dropped support for Qt4. The last release with Qt4 '
        'support was v2021.09.14.')

qt5_backend_available = 'Qt5Agg' in matplotlib.rcsetup.all_backends

if not qt5_backend_available:
    logger.warn(
        'Matplotlib Qt5Agg backend is not available. Snufflings drawing '
        'matplotlib figures may not work properly.')

if matplotlib.get_backend().find('Qt4') != -1:
    matplotlib.use('Qt5Agg')

import PyQt5 as PyQt  # noqa
from PyQt5 import Qt  # noqa
from PyQt5 import QtCore as qc  # noqa
from PyQt5 import QtGui as qg  # noqa
from PyQt5 import QtWidgets as qw  # noqa
from PyQt5 import QtNetwork as qn  # noqa
from PyQt5 import QtSvg as qsvg  # noqa
from PyQt5 import QtPrintSupport as qprint  # noqa

def getSaveFileName(*args, **kwargs):
     return qw.QFileDialog.getSaveFileName(*args, **kwargs)[0]

class QPixmapCache(qg.QPixmapCache):
    def cached(self, key):
        return self.find(key)

try:
    vers = qc.QVersionNumber.fromString
except AttributeError:
    def vers(s):
        return tuple(s.split('.'))

# Application attribute has to be set for QWebView
if vers(Qt.QT_VERSION_STR) >= vers('5.4.0'):
    Qt.QCoreApplication.setAttribute(qc.Qt.AA_ShareOpenGLContexts, True)


def get_em(painter):
    return painter.boundingRect(qc.QRectF(), 0, 'M').width()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

'''
Data model and content types handled by the Squirrel framework.

Squirrel uses flat content types to represent waveform, station, channel,
response, event, and a few other objects. A common subset of the information in
these objects is indexed in the database, currently: kind, codes, time interval
and sampling rate. The :py:class:`Nut` objects encapsulate this information
together with information about where and how to get the associated bulk data.

Further content types are defined here to handle waveform orders, waveform
promises, data coverage and sensor information.
'''

import re
import fnmatch
import math
import hashlib
from os import urandom
from base64 import urlsafe_b64encode
from collections import defaultdict, namedtuple

import numpy as num

from pyrockoeost import util
from pyrockoeost.guts import Object, SObject, String, Timestamp, Float, Int, \
    Unicode, Tuple, List, StringChoice, Any, Dict, clone
from pyrockoeost.model import squirrel_content, Location
from pyrockoeost.response import FrequencyResponse, MultiplyResponse, \
    IntegrationResponse, DifferentiationResponse, simplify_responses, \
    FrequencyResponseCheckpoint

from .error import ConversionError, SquirrelError

d2r = num.pi / 180.
r2d = 1.0 / d2r


guts_prefix = 'squirrel'


def mkvec(x, y, z):
    return num.array([x, y, z], dtype=float)


def are_orthogonal(vecs, eps=0.01):
    return all(abs(
        num.dot(vecs[i], vecs[j]) < eps
        for (i, j) in [(0, 1), (1, 2), (2, 0)]))


g_codes_pool = {}


class CodesError(SquirrelError):
    pass


class Codes(SObject):
    pass


def normalize_nslce(*args, **kwargs):
    if args and kwargs:
        raise ValueError('Either *args or **kwargs accepted, not both.')

    if len(args) == 1:
        if isinstance(args[0], str):
            args = tuple(args[0].split('.'))
        elif isinstance(args[0], tuple):
            args = args[0]
        else:
            raise ValueError('Invalid argument type: %s' % type(args[0]))

    nargs = len(args)
    if nargs == 5:
        t = args

    elif nargs == 4:
        t = args + ('',)

    elif nargs == 0:
        d = dict(
            network='',
            station='',
            location='',
            channel='',
            extra='')

        d.update(kwargs)
        t = tuple(kwargs.get(k, '') for k in (
            'network', 'station', 'location', 'channel', 'extra'))

    else:
        raise CodesError(
            'Does not match NSLC or NSLCE codes pattern: %s' % '.'.join(args))

    if '.'.join(t).count('.') != 4:
        raise CodesError(
            'Codes may not contain a ".": "%s", "%s", "%s", "%s", "%s"' % t)

    return t


CodesNSLCEBase = namedtuple(
    'CodesNSLCEBase', [
        'network', 'station', 'location', 'channel', 'extra'])


class CodesNSLCE(CodesNSLCEBase, Codes):
    '''
    Codes denominating a seismic channel (NSLC or NSLCE).

    FDSN/SEED style NET.STA.LOC.CHA is accepted or NET.STA.LOC.CHA.EXTRA, where
    the EXTRA part in the latter form can be used to identify a custom
    processing applied to a channel.
    '''

    __slots__ = ()
    __hash__ = CodesNSLCEBase.__hash__

    as_dict = CodesNSLCEBase._asdict

    def __new__(cls, *args, safe_str=None, **kwargs):
        nargs = len(args)
        if nargs == 1 and isinstance(args[0], CodesNSLCE):
            return args[0]
        elif nargs == 1 and isinstance(args[0], CodesNSL):
            t = (args[0].nsl) + ('*', '*')
        elif nargs == 1 and isinstance(args[0], CodesX):
            t = ('*', '*', '*', '*', '*')
        elif safe_str is not None:
            t = safe_str.split('.')
        else:
            t = normalize_nslce(*args, **kwargs)

        x = CodesNSLCEBase.__new__(cls, *t)
        return g_codes_pool.setdefault(x, x)

    def __init__(self, *args, **kwargs):
        Codes.__init__(self)

    def __str__(self):
        if self.extra == '':
            return '.'.join(self[:-1])
        else:
            return '.'.join(self)

    def __eq__(self, other):
        if not isinstance(other, CodesNSLCE):
            other = CodesNSLCE(other)

        return CodesNSLCEBase.__eq__(self, other)

    def matches(self, pattern):
        if not isinstance(pattern, CodesNSLCE):
            pattern = CodesNSLCE(pattern)

        return match_codes(pattern, self)

    @property
    def safe_str(self):
        return '.'.join(self)

    @property
    def nslce(self):
        return self[:5]

    @property
    def nslc(self):
        return self[:4]

    @property
    def nsl(self):
        return self[:3]

    @property
    def ns(self):
        return self[:2]

    @property
    def codes_nsl(self):
        return CodesNSL(self)

    @property
    def codes_nsl_star(self):
        return CodesNSL(self.network, self.station, '*')

    def as_tuple(self):
        return tuple(self)

    def replace(self, **kwargs):
        x = CodesNSLCEBase._replace(self, **kwargs)
        return g_codes_pool.setdefault(x, x)


def normalize_nsl(*args, **kwargs):
    if args and kwargs:
        raise ValueError('Either *args or **kwargs accepted, not both.')

    if len(args) == 1:
        if isinstance(args[0], str):
            args = tuple(args[0].split('.'))
        elif isinstance(args[0], tuple):
            args = args[0]
        else:
            raise ValueError('Invalid argument type: %s' % type(args[0]))

    nargs = len(args)
    if nargs == 3:
        t = args

    elif nargs == 0:
        d = dict(
            network='',
            station='',
            location='')

        d.update(kwargs)
        t = tuple(kwargs.get(k, '') for k in (
            'network', 'station', 'location'))

    else:
        raise CodesError(
            'Does not match NSL codes pattern: %s' % '.'.join(args))

    if '.'.join(t).count('.') != 2:
        raise CodesError(
            'Codes may not contain a ".": "%s", "%s", "%s"' % t)

    return t


CodesNSLBase = namedtuple(
    'CodesNSLBase', [
        'network', 'station', 'location'])


class CodesNSL(CodesNSLBase, Codes):
    '''
    Codes denominating a seismic station (NSL).

    NET.STA.LOC is accepted, slightly different from SEED/StationXML, where
    LOC is part of the channel. By setting location='*' is possible to get
    compatible behaviour in most cases.
    '''

    __slots__ = ()
    __hash__ = CodesNSLBase.__hash__

    as_dict = CodesNSLBase._asdict

    def __new__(cls, *args, safe_str=None, **kwargs):
        nargs = len(args)
        if nargs == 1 and isinstance(args[0], CodesNSL):
            return args[0]
        elif nargs == 1 and isinstance(args[0], CodesNSLCE):
            t = args[0].nsl
        elif nargs == 1 and isinstance(args[0], CodesX):
            t = ('*', '*', '*')
        elif safe_str is not None:
            t = safe_str.split('.')
        else:
            t = normalize_nsl(*args, **kwargs)

        x = CodesNSLBase.__new__(cls, *t)
        return g_codes_pool.setdefault(x, x)

    def __init__(self, *args, **kwargs):
        Codes.__init__(self)

    def __str__(self):
        return '.'.join(self)

    def __eq__(self, other):
        if not isinstance(other, CodesNSL):
            other = CodesNSL(other)

        return CodesNSLBase.__eq__(self, other)

    def matches(self, pattern):
        if not isinstance(pattern, CodesNSL):
            pattern = CodesNSL(pattern)

        return match_codes(pattern, self)

    @property
    def safe_str(self):
        return '.'.join(self)

    @property
    def ns(self):
        return self[:2]

    @property
    def nsl(self):
        return self[:3]

    def as_tuple(self):
        return tuple(self)

    def replace(self, **kwargs):
        x = CodesNSLBase._replace(self, **kwargs)
        return g_codes_pool.setdefault(x, x)


CodesXBase = namedtuple(
    'CodesXBase', [
        'name'])


class CodesX(CodesXBase, Codes):
    '''
    General purpose codes for anything other than channels or stations.
    '''

    __slots__ = ()
    __hash__ = CodesXBase.__hash__
    __eq__ = CodesXBase.__eq__

    as_dict = CodesXBase._asdict

    def __new__(cls, name='', safe_str=None):
        if isinstance(name, CodesX):
            return name
        elif isinstance(name, (CodesNSLCE, CodesNSL)):
            name = '*'
        elif safe_str is not None:
            name = safe_str
        else:
            if '.' in name:
                raise CodesError('Code may not contain a ".": %s' % name)

        x = CodesXBase.__new__(cls, name)
        return g_codes_pool.setdefault(x, x)

    def __init__(self, *args, **kwargs):
        Codes.__init__(self)

    def __str__(self):
        return '.'.join(self)

    @property
    def safe_str(self):
        return '.'.join(self)

    @property
    def ns(self):
        return self[:2]

    def as_tuple(self):
        return tuple(self)

    def replace(self, **kwargs):
        x = CodesXBase._replace(self, **kwargs)
        return g_codes_pool.setdefault(x, x)


g_codes_patterns = {}


def match_codes(pattern, codes):
    spattern = pattern.safe_str
    scodes = codes.safe_str
    if spattern not in g_codes_patterns:
        rpattern = re.compile(fnmatch.translate(spattern), re.I)
        g_codes_patterns[spattern] = rpattern

    rpattern = g_codes_patterns[spattern]
    return bool(rpattern.match(scodes))


g_content_kinds = [
    'undefined',
    'waveform',
    'station',
    'channel',
    'response',
    'event',
    'waveform_promise']


g_codes_classes = [
    CodesX,
    CodesNSLCE,
    CodesNSL,
    CodesNSLCE,
    CodesNSLCE,
    CodesX,
    CodesNSLCE]

g_codes_classes_ndot = {
    0: CodesX,
    2: CodesNSL,
    3: CodesNSLCE,
    4: CodesNSLCE}


def to_codes_simple(kind_id, codes_safe_str):
    return g_codes_classes[kind_id](safe_str=codes_safe_str)


def to_codes(kind_id, obj):
    return g_codes_classes[kind_id](obj)


def to_codes_guess(s):
    try:
        return g_codes_classes_ndot[s.count('.')](s)
    except KeyError:
        raise CodesError('Cannot guess codes type: %s' % s)


# derived list class to enable detection of already preprocessed codes patterns
class codes_patterns_list(list):
    pass


def codes_patterns_for_kind(kind_id, codes):
    if isinstance(codes, codes_patterns_list):
        return codes

    if isinstance(codes, list):
        lcodes = codes_patterns_list()
        for sc in codes:
            lcodes.extend(codes_patterns_for_kind(kind_id, sc))

        return lcodes

    codes = to_codes(kind_id, codes)

    lcodes = codes_patterns_list()
    lcodes.append(codes)
    if kind_id == STATION:
        lcodes.append(codes.replace(location='[*]'))

    return lcodes


g_content_kind_ids = (
    UNDEFINED, WAVEFORM, STATION, CHANNEL, RESPONSE, EVENT,
    WAVEFORM_PROMISE) = range(len(g_content_kinds))


g_tmin, g_tmax = util.get_working_system_time_range()[:2]


try:
    g_tmin_queries = max(g_tmin, util.str_to_time_fillup('1900-01-01'))
except Exception:
    g_tmin_queries = g_tmin


def to_kind(kind_id):
    return g_content_kinds[kind_id]


def to_kinds(kind_ids):
    return [g_content_kinds[kind_id] for kind_id in kind_ids]


def to_kind_id(kind):
    return g_content_kinds.index(kind)


def to_kind_ids(kinds):
    return [g_content_kinds.index(kind) for kind in kinds]


g_kind_mask_all = 0xff


def to_kind_mask(kinds):
    if kinds:
        return sum(1 << kind_id for kind_id in to_kind_ids(kinds))
    else:
        return g_kind_mask_all


def str_or_none(x):
    if x is None:
        return None
    else:
        return str(x)


def float_or_none(x):
    if x is None:
        return None
    else:
        return float(x)


def int_or_none(x):
    if x is None:
        return None
    else:
        return int(x)


def int_or_g_tmin(x):
    if x is None:
        return g_tmin
    else:
        return int(x)


def int_or_g_tmax(x):
    if x is None:
        return g_tmax
    else:
        return int(x)


def tmin_or_none(x):
    if x == g_tmin:
        return None
    else:
        return x


def tmax_or_none(x):
    if x == g_tmax:
        return None
    else:
        return x


def time_or_none_to_str(x):
    if x is None:
        return '...'.ljust(17)
    else:
        return util.time_to_str(x)


def codes_to_str_abbreviated(codes, indent='  '):
    codes = [str(x) for x in codes]

    if len(codes) > 20:
        scodes = '\n' + util.ewrap(codes[:10], indent=indent) \
            + '\n%s[%i more]\n' % (indent, len(codes) - 20) \
            + util.ewrap(codes[-10:], indent='  ')
    else:
        scodes = '\n' + util.ewrap(codes, indent=indent) \
            if codes else '<none>'

    return scodes


g_offset_time_unit_inv = 1000000000
g_offset_time_unit = 1.0 / g_offset_time_unit_inv


def tsplit(t):
    if t is None:
        return None, 0

    t = util.to_time_float(t)
    if type(t) is float:
        t = round(t, 5)
    else:
        t = round(t, 9)

    seconds = num.floor(t)
    offset = t - seconds
    return int(seconds), int(round(offset * g_offset_time_unit_inv))


def tjoin(seconds, offset):
    if seconds is None:
        return None

    return util.to_time_float(seconds) \
        + util.to_time_float(offset*g_offset_time_unit)


tscale_min = 1
tscale_max = 365 * 24 * 3600  # last edge is one above
tscale_logbase = 20

tscale_edges = [tscale_min]
while True:
    tscale_edges.append(tscale_edges[-1]*tscale_logbase)
    if tscale_edges[-1] >= tscale_max:
        break


tscale_edges = num.array(tscale_edges)


def tscale_to_kscale(tscale):

    # 0 <= x < tscale_edges[1]: 0
    # tscale_edges[1] <= x < tscale_edges[2]: 1
    # ...
    # tscale_edges[len(tscale_edges)-1] <= x: len(tscale_edges)

    return int(num.searchsorted(tscale_edges, tscale))


@squirrel_content
class Station(Location):
    '''
    A seismic station.
    '''

    codes = CodesNSL.T()

    tmin = Timestamp.T(optional=True)
    tmax = Timestamp.T(optional=True)

    description = Unicode.T(optional=True)

    def __init__(self, **kwargs):
        kwargs['codes'] = CodesNSL(kwargs['codes'])
        Location.__init__(self, **kwargs)

    @property
    def time_span(self):
        return (self.tmin, self.tmax)

    def get_pyrockoeost_station(self):
        from pyrockoeost import model
        return model.Station(*self._get_pyrockoeost_station_args())

    def _get_pyrockoeost_station_args(self):
        return (
            self.codes.network,
            self.codes.station,
            self.codes.location,
            self.lat,
            self.lon,
            self.elevation,
            self.depth,
            self.north_shift,
            self.east_shift)


@squirrel_content
class ChannelBase(Location):
    codes = CodesNSLCE.T()

    tmin = Timestamp.T(optional=True)
    tmax = Timestamp.T(optional=True)

    deltat = Float.T(optional=True)

    def __init__(self, **kwargs):
        kwargs['codes'] = CodesNSLCE(kwargs['codes'])
        Location.__init__(self, **kwargs)

    @property
    def time_span(self):
        return (self.tmin, self.tmax)

    def _get_sensor_codes(self):
        return self.codes.replace(
            channel=self.codes.channel[:-1] + '?')

    def _get_sensor_args(self):
        def getattr_rep(k):
            if k == 'codes':
                return self._get_sensor_codes()
            else:
                return getattr(self, k)

        return tuple(getattr_rep(k) for k in ChannelBase.T.propnames)

    def _get_channel_args(self, component):
        def getattr_rep(k):
            if k == 'codes':
                return self.codes.replace(
                    channel=self.codes.channel[:-1] + component)
            else:
                return getattr(self, k)

        return tuple(getattr_rep(k) for k in ChannelBase.T.propnames)

    def _get_pyrockoeost_station_args(self):
        return (
            self.codes.network,
            self.codes.station,
            self.codes.location,
            self.lat,
            self.lon,
            self.elevation,
            self.depth,
            self.north_shift,
            self.east_shift)


class Channel(ChannelBase):
    '''
    A channel of a seismic station.
    '''

    dip = Float.T(optional=True)
    azimuth = Float.T(optional=True)

    def get_pyrockoeost_channel(self):
        from pyrockoeost import model
        return model.Channel(*self._get_pyrockoeost_channel_args())

    def _get_pyrockoeost_channel_args(self):
        return (
            self.codes.channel,
            self.azimuth,
            self.dip)

    @property
    def orientation_enz(self):
        if None in (self.azimuth, self.dip):
            return None

        n = math.cos(self.azimuth*d2r)*math.cos(self.dip*d2r)
        e = math.sin(self.azimuth*d2r)*math.cos(self.dip*d2r)
        d = math.sin(self.dip*d2r)
        return mkvec(e, n, -d)


def cut_intervals(channels):
    channels = list(channels)
    times = set()
    for channel in channels:
        if channel.tmin is not None:
            times.add(channel.tmin)
        if channel.tmax is not None:
            times.add(channel.tmax)

    times = sorted(times)

    if any(channel.tmin is None for channel in channels):
        times[0:0] = [None]

    if any(channel.tmax is None for channel in channels):
        times.append(None)

    if len(times) <= 2:
        return channels

    channels_out = []
    for channel in channels:
        for i in range(len(times)-1):
            tmin = times[i]
            tmax = times[i+1]
            if ((channel.tmin is None or (
                    tmin is not None and channel.tmin <= tmin))
                    and (channel.tmax is None or (
                        tmax is not None and tmax <= channel.tmax))):

                channel_out = clone(channel)
                channel_out.tmin = tmin
                channel_out.tmax = tmax
                channels_out.append(channel_out)

    return channels_out


class Sensor(ChannelBase):
    '''
    Representation of a channel group.
    '''

    channels = List.T(Channel.T())

    @classmethod
    def from_channels(cls, channels):
        groups = defaultdict(list)
        for channel in channels:
            groups[channel._get_sensor_codes()].append(channel)

        channels_cut = []
        for group in groups.values():
            channels_cut.extend(cut_intervals(group))

        groups = defaultdict(list)
        for channel in channels_cut:
            groups[channel._get_sensor_args()].append(channel)

        return [
            cls(channels=channels,
                **dict(zip(ChannelBase.T.propnames, args)))
            for args, channels in groups.items()]

    def channel_vectors(self):
        return num.vstack(
            [channel.orientation_enz for channel in self.channels])

    def projected_channels(self, system, component_names):
        return [
            Channel(
                azimuth=math.atan2(e, n) * r2d,
                dip=-math.asin(z) * r2d,
                **dict(zip(
                    ChannelBase.T.propnames,
                    self._get_channel_args(comp))))
            for comp, (e, n, z) in zip(component_names, system)]

    def matrix_to(self, system, epsilon=1e-7):
        m = num.dot(system, self.channel_vectors().T)
        m[num.abs(m) < epsilon] = 0.0
        return m

    def projection_to(self, system, component_names):
        return (
            self.matrix_to(system),
            self.channels,
            self.projected_channels(system, component_names))

    def projection_to_enz(self):
        return self.projection_to(num.identity(3), 'ENZ')

    def projection_to_trz(self, source, azimuth=None):
        if azimuth is not None:
            assert source is None
        else:
            azimuth = source.azibazi_to(self)[1] + 180.

        return self.projection_to(num.array([
            [math.cos(azimuth*d2r), -math.sin(azimuth*d2r), 0.],
            [math.sin(azimuth*d2r), math.cos(azimuth*d2r), 0.],
            [0., 0., 1.]], dtype=float), 'TRZ')

    def project_to_enz(self, traces):
        from pyrockoeost import trace

        matrix, in_channels, out_channels = self.projection_to_enz()

        return trace.project(traces, matrix, in_channels, out_channels)

    def project_to_trz(self, source, traces, azimuth=None):
        from pyrockoeost import trace

        matrix, in_channels, out_channels = self.projection_to_trz(
            source, azimuth=azimuth)

        return trace.project(traces, matrix, in_channels, out_channels)


observational_quantities = [
    'acceleration', 'velocity', 'displacement', 'pressure',
    'rotation-displacement', 'rotation-velocity', 'rotation-acceleration',
    'temperature']


technical_quantities = [
    'voltage', 'counts']


class QuantityType(StringChoice):
    '''
    Choice of observational or technical quantity.

    SI units are used for all quantities, where applicable.
    '''
    choices = observational_quantities + technical_quantities


class ResponseStage(Object):
    '''
    Representation of a response stage.

    Components of a seismic recording system are represented as a sequence of
    response stages, e.g. sensor, pre-amplifier, digitizer, digital
    downsampling.
    '''
    input_quantity = QuantityType.T(optional=True)
    input_sample_rate = Float.T(optional=True)
    output_quantity = QuantityType.T(optional=True)
    output_sample_rate = Float.T(optional=True)
    elements = List.T(FrequencyResponse.T())
    log = List.T(Tuple.T(3, String.T()))

    @property
    def stage_type(self):
        if self.input_quantity in observational_quantities \
                and self.output_quantity in observational_quantities:
            return 'conversion'

        if self.input_quantity in observational_quantities \
                and self.output_quantity == 'voltage':
            return 'sensor'

        elif self.input_quantity == 'voltage' \
                and self.output_quantity == 'voltage':
            return 'electronics'

        elif self.input_quantity == 'voltage' \
                and self.output_quantity == 'counts':
            return 'digitizer'

        elif self.decimation_factor is not None \
                and (self.input_quantity is None or self.input_quantity == 'counts') \
                and (self.output_quantity is None or self.output_quantity == 'counts'):  # noqa
            return 'decimation'

        elif self.input_quantity in observational_quantities \
                and self.output_quantity == 'counts':
            return 'combination'

        else:
            return 'unknown'

    @property
    def decimation_factor(self):
        irate = self.input_sample_rate
        orate = self.output_sample_rate
        if irate is not None and orate is not None \
                and irate > orate and irate / orate > 1.0:

            return irate / orate
        else:
            return None

    @property
    def summary_quantities(self):
        return '%s -> %s' % (
            self.input_quantity or '?',
            self.output_quantity or '?')

    @property
    def summary_rates(self):
        irate = self.input_sample_rate
        orate = self.output_sample_rate
        factor = self.decimation_factor

        if irate and orate is None:
            return '%g Hz' % irate

        elif orate and irate is None:
            return '%g Hz' % orate

        elif irate and orate and irate == orate:
            return '%g Hz' % irate

        elif any(x for x in (irate, orate, factor)):
            return '%s -> %s Hz (%s)' % (
                '%g' % irate if irate else '?',
                '%g' % orate if orate else '?',
                ':%g' % factor if factor else '?')
        else:
            return ''

    @property
    def summary_elements(self):
        xs = [x.summary for x in self.elements]
        return '%s' % ('*'.join(x for x in xs if x != 'one') or 'one')

    @property
    def summary_log(self):
        return ''.join(sorted(set(x[0][0].upper() for x in self.log)))

    @property
    def summary_entries(self):
        return (
            self.__class__.__name__,
            self.stage_type,
            self.summary_log,
            self.summary_quantities,
            self.summary_rates,
            self.summary_elements)

    @property
    def summary(self):
        return util.fmt_summary(self.summary_entries, (10, 15, 3, 30, 30, 0))

    def get_effective(self):
        return MultiplyResponse(responses=list(self.elements))


D = 'displacement'
V = 'velocity'
A = 'acceleration'

g_converters = {
    (V, D): IntegrationResponse(1),
    (A, D): IntegrationResponse(2),
    (D, V): DifferentiationResponse(1),
    (A, V): IntegrationResponse(1),
    (D, A): DifferentiationResponse(2),
    (V, A): DifferentiationResponse(1)}


def response_converters(input_quantity, output_quantity):
    if input_quantity is None or input_quantity == output_quantity:
        return []

    if output_quantity is None:
        raise ConversionError('Unspecified target quantity.')

    try:
        return [g_converters[input_quantity, output_quantity]]

    except KeyError:
        raise ConversionError('No rule to convert from "%s" to "%s".' % (
            input_quantity, output_quantity))


@squirrel_content
class Response(Object):
    '''
    The instrument response of a seismic station channel.
    '''

    codes = CodesNSLCE.T()
    tmin = Timestamp.T(optional=True)
    tmax = Timestamp.T(optional=True)

    stages = List.T(ResponseStage.T())
    checkpoints = List.T(FrequencyResponseCheckpoint.T())

    deltat = Float.T(optional=True)
    log = List.T(Tuple.T(3, String.T()))

    def __init__(self, **kwargs):
        kwargs['codes'] = CodesNSLCE(kwargs['codes'])
        Object.__init__(self, **kwargs)

    @property
    def time_span(self):
        return (self.tmin, self.tmax)

    @property
    def nstages(self):
        return len(self.stages)

    @property
    def input_quantity(self):
        return self.stages[0].input_quantity if self.stages else None

    @property
    def output_quantity(self):
        return self.stages[-1].output_quantity if self.stages else None

    @property
    def output_sample_rate(self):
        return self.stages[-1].output_sample_rate if self.stages else None

    @property
    def summary_stages(self):
        def grouped(xs):
            xs = list(xs)
            g = []
            for i in range(len(xs)):
                g.append(xs[i])
                if i+1 < len(xs) and xs[i+1] != xs[i]:
                    yield g
                    g = []

            if g:
                yield g

        return '+'.join(
            '%s%s' % (g[0], '(%i)' % len(g) if len(g) > 1 else '')
            for g in grouped(stage.stage_type for stage in self.stages))

    @property
    def summary_quantities(self):
        orate = self.output_sample_rate
        return '%s -> %s%s' % (
            self.input_quantity or '?',
            self.output_quantity or '?',
            ' @ %g Hz' % orate if orate else '')

    @property
    def summary_log(self):
        return ''.join(sorted(set(x[0][0].upper() for x in self.log)))

    @property
    def summary_entries(self):
        return (
            self.__class__.__name__,
            str(self.codes),
            self.str_time_span,
            self.summary_log,
            self.summary_quantities,
            self.summary_stages)

    @property
    def summary(self):
        return util.fmt_summary(self.summary_entries, (10, 20, 55, 3, 35, 0))

    def get_effective(self, input_quantity=None, stages=(None, None)):
        try:
            elements = response_converters(input_quantity, self.input_quantity)
        except ConversionError as e:
            raise ConversionError(str(e) + ' (%s)' % self.summary)

        elements.extend(
            stage.get_effective() for stage in self.stages[slice(*stages)])

        if input_quantity is None \
                or input_quantity == self.input_quantity:
            checkpoints = self.checkpoints
        else:
            checkpoints = []

        return MultiplyResponse(
            responses=simplify_responses(elements),
            checkpoints=checkpoints)


@squirrel_content
class Event(Object):
    '''
    A seismic event.
    '''

    name = String.T(optional=True)
    time = Timestamp.T()
    duration = Float.T(optional=True)

    lat = Float.T()
    lon = Float.T()
    depth = Float.T(optional=True)

    magnitude = Float.T(optional=True)

    def get_pyrockoeost_event(self):
        from pyrockoeost import model
        model.Event(
            name=self.name,
            time=self.time,
            lat=self.lat,
            lon=self.lon,
            depth=self.depth,
            magnitude=self.magnitude,
            duration=self.duration)

    @property
    def time_span(self):
        return (self.time, self.time)


def ehash(s):
    return hashlib.sha1(s.encode('utf8')).hexdigest()


def random_name(n=8):
    return urlsafe_b64encode(urandom(n)).rstrip(b'=').decode('ascii')


@squirrel_content
class WaveformPromise(Object):
    '''
    Information about a waveform potentially downloadable from a remote site.

    In the Squirrel framework, waveform promises are used to permit download of
    selected waveforms from a remote site. They are typically generated by
    calls to
    :py:meth:`~pyrockoeost.squirrel.base.Squirrel.update_waveform_promises`.
    Waveform promises are inserted and indexed in the database similar to
    normal waveforms. When processing a waveform query, e.g. from
    :py:meth:`~pyrockoeost.squirrel.base.Squirrel.get_waveform`, and no local
    waveform is available for the queried time span, a matching promise can be
    resolved, i.e. an attempt is made to download the waveform from the remote
    site. The promise is removed after the download attempt (except when a
    network error occurs). This prevents Squirrel from making unnecessary
    queries for waveforms missing at the remote site.
    '''

    codes = CodesNSLCE.T()
    tmin = Timestamp.T()
    tmax = Timestamp.T()

    deltat = Float.T(optional=True)

    source_hash = String.T()

    def __init__(self, **kwargs):
        kwargs['codes'] = CodesNSLCE(kwargs['codes'])
        Object.__init__(self, **kwargs)

    @property
    def time_span(self):
        return (self.tmin, self.tmax)


class InvalidWaveform(Exception):
    pass


class WaveformOrder(Object):
    '''
    Waveform request information.
    '''

    source_id = String.T()
    codes = CodesNSLCE.T()
    deltat = Float.T()
    tmin = Timestamp.T()
    tmax = Timestamp.T()
    gaps = List.T(Tuple.T(2, Timestamp.T()))

    @property
    def client(self):
        return self.source_id.split(':')[1]

    def describe(self, site='?'):
        return '%s:%s %s [%s - %s]' % (
            self.client, site, str(self.codes),
            util.time_to_str(self.tmin), util.time_to_str(self.tmax))

    def validate(self, tr):
        if tr.ydata.size == 0:
            raise InvalidWaveform(
                'waveform with zero data samples')

        if tr.deltat != self.deltat:
            raise InvalidWaveform(
                'incorrect sampling interval - waveform: %g s, '
                'meta-data: %g s' % (
                    tr.deltat, self.deltat))

        if not num.all(num.isfinite(tr.ydata)):
            raise InvalidWaveform('waveform has NaN values')

    def estimate_nsamples(self):
        return int(round((self.tmax - self.tmin) / self.deltat))+1


def order_summary(orders):
    codes_list = sorted(set(order.codes for order in orders))
    if len(codes_list) > 3:
        return '%i order%s: %s - %s' % (
            len(orders),
            util.plural_s(orders),
            str(codes_list[0]),
            str(codes_list[1]))

    else:
        return '%i order%s: %s' % (
            len(orders),
            util.plural_s(orders),
            ', '.join(str(codes) for codes in codes_list))


class Nut(Object):
    '''
    Index entry referencing an elementary piece of content.

    So-called *nuts* are used in pyrockoeost's Squirrel framework to hold common
    meta-information about individual pieces of waveforms, stations, channels,
    etc. together with the information where it was found or generated.
    '''

    file_path = String.T(optional=True)
    file_format = String.T(optional=True)
    file_mtime = Timestamp.T(optional=True)
    file_size = Int.T(optional=True)

    file_segment = Int.T(optional=True)
    file_element = Int.T(optional=True)

    kind_id = Int.T()
    codes = Codes.T()

    tmin_seconds = Int.T(default=0)
    tmin_offset = Int.T(default=0, optional=True)
    tmax_seconds = Int.T(default=0)
    tmax_offset = Int.T(default=0, optional=True)

    deltat = Float.T(default=0.0)

    content = Any.T(optional=True)
    raw_content = Dict.T(String.T(), Any.T())

    content_in_db = False

    def __init__(
            self,
            file_path=None,
            file_format=None,
            file_mtime=None,
            file_size=None,
            file_segment=None,
            file_element=None,
            kind_id=0,
            codes=CodesX(''),
            tmin_seconds=None,
            tmin_offset=0,
            tmax_seconds=None,
            tmax_offset=0,
            deltat=None,
            content=None,
            raw_content=None,
            tmin=None,
            tmax=None,
            values_nocheck=None):

        if values_nocheck is not None:
            (self.file_path, self.file_format, self.file_mtime,
             self.file_size,
             self.file_segment, self.file_element,
             self.kind_id, codes_safe_str,
             self.tmin_seconds, self.tmin_offset,
             self.tmax_seconds, self.tmax_offset,
             self.deltat) = values_nocheck

            self.codes = to_codes_simple(self.kind_id, codes_safe_str)
            self.deltat = self.deltat or None
            self.raw_content = {}
            self.content = None
        else:
            if tmin is not None:
                tmin_seconds, tmin_offset = tsplit(tmin)

            if tmax is not None:
                tmax_seconds, tmax_offset = tsplit(tmax)

            self.kind_id = int(kind_id)
            self.codes = codes
            self.tmin_seconds = int_or_g_tmin(tmin_seconds)
            self.tmin_offset = int(tmin_offset)
            self.tmax_seconds = int_or_g_tmax(tmax_seconds)
            self.tmax_offset = int(tmax_offset)
            self.deltat = float_or_none(deltat)
            self.file_path = str_or_none(file_path)
            self.file_segment = int_or_none(file_segment)
            self.file_element = int_or_none(file_element)
            self.file_format = str_or_none(file_format)
            self.file_mtime = float_or_none(file_mtime)
            self.file_size = int_or_none(file_size)
            self.content = content
            if raw_content is None:
                self.raw_content = {}
            else:
                self.raw_content = raw_content

        Object.__init__(self, init_props=False)

    def __eq__(self, other):
        return (isinstance(other, Nut) and
                self.equality_values == other.equality_values)

    def hash(self):
        return ehash(','.join(str(x) for x in self.key))

    def __ne__(self, other):
        return not (self == other)

    def get_io_backend(self):
        from . import io
        return io.get_backend(self.file_format)

    def file_modified(self):
        return self.get_io_backend().get_stats(self.file_path) \
            != (self.file_mtime, self.file_size)

    @property
    def dkey(self):
        return (self.kind_id, self.tmin_seconds, self.tmin_offset, self.codes)

    @property
    def key(self):
        return (
            self.file_path,
            self.file_segment,
            self.file_element,
            self.file_mtime)

    @property
    def equality_values(self):
        return (
            self.file_segment, self.file_element,
            self.kind_id, self.codes,
            self.tmin_seconds, self.tmin_offset,
            self.tmax_seconds, self.tmax_offset, self.deltat)

    def diff(self, other):
        names = [
            'file_segment', 'file_element', 'kind_id', 'codes',
            'tmin_seconds', 'tmin_offset', 'tmax_seconds', 'tmax_offset',
            'deltat']

        d = []
        for name, a, b in zip(
                names, self.equality_values, other.equality_values):

            if a != b:
                d.append((name, a, b))

        return d

    @property
    def tmin(self):
        return tjoin(self.tmin_seconds, self.tmin_offset)

    @tmin.setter
    def tmin(self, t):
        self.tmin_seconds, self.tmin_offset = tsplit(t)

    @property
    def tmax(self):
        return tjoin(self.tmax_seconds, self.tmax_offset)

    @tmax.setter
    def tmax(self, t):
        self.tmax_seconds, self.tmax_offset = tsplit(t)

    @property
    def kscale(self):
        if self.tmin_seconds is None or self.tmax_seconds is None:
            return 0
        return tscale_to_kscale(self.tmax_seconds - self.tmin_seconds)

    @property
    def waveform_kwargs(self):
        network, station, location, channel, extra = self.codes

        return dict(
            network=network,
            station=station,
            location=location,
            channel=channel,
            extra=extra,
            tmin=self.tmin,
            tmax=self.tmax,
            deltat=self.deltat)

    @property
    def waveform_promise_kwargs(self):
        return dict(
            codes=self.codes,
            tmin=self.tmin,
            tmax=self.tmax,
            deltat=self.deltat)

    @property
    def station_kwargs(self):
        network, station, location = self.codes
        return dict(
            codes=self.codes,
            tmin=tmin_or_none(self.tmin),
            tmax=tmax_or_none(self.tmax))

    @property
    def channel_kwargs(self):
        network, station, location, channel, extra = self.codes
        return dict(
            codes=self.codes,
            tmin=tmin_or_none(self.tmin),
            tmax=tmax_or_none(self.tmax),
            deltat=self.deltat)

    @property
    def response_kwargs(self):
        return dict(
            codes=self.codes,
            tmin=tmin_or_none(self.tmin),
            tmax=tmax_or_none(self.tmax),
            deltat=self.deltat)

    @property
    def event_kwargs(self):
        return dict(
            name=self.codes,
            time=self.tmin,
            duration=(self.tmax - self.tmin) or None)

    @property
    def trace_kwargs(self):
        network, station, location, channel, extra = self.codes

        return dict(
            network=network,
            station=station,
            location=location,
            channel=channel,
            extra=extra,
            tmin=self.tmin,
            tmax=self.tmax-self.deltat,
            deltat=self.deltat)

    @property
    def dummy_trace(self):
        return DummyTrace(self)

    @property
    def summary_entries(self):
        if self.tmin == self.tmax:
            ts = util.time_to_str(self.tmin)
        else:
            ts = '%s - %s' % (
                util.time_to_str(self.tmin),
                util.time_to_str(self.tmax))

        return (
            self.__class__.__name__,
            to_kind(self.kind_id),
            str(self.codes),
            ts)

    @property
    def summary(self):
        return util.fmt_summary(self.summary_entries, (10, 16, 20, 0))


def make_waveform_nut(**kwargs):
    return Nut(kind_id=WAVEFORM, **kwargs)


def make_waveform_promise_nut(**kwargs):
    return Nut(kind_id=WAVEFORM_PROMISE, **kwargs)


def make_station_nut(**kwargs):
    return Nut(kind_id=STATION, **kwargs)


def make_channel_nut(**kwargs):
    return Nut(kind_id=CHANNEL, **kwargs)


def make_response_nut(**kwargs):
    return Nut(kind_id=RESPONSE, **kwargs)


def make_event_nut(**kwargs):
    return Nut(kind_id=EVENT, **kwargs)


def group_channels(nuts):
    by_ansl = {}
    for nut in nuts:
        if nut.kind_id != CHANNEL:
            continue

        ansl = nut.codes[:4]

        if ansl not in by_ansl:
            by_ansl[ansl] = {}

        group = by_ansl[ansl]

        k = nut.codes[4][:-1], nut.deltat, nut.tmin, nut.tmax

        if k not in group:
            group[k] = set()

        group.add(nut.codes[4])

    return by_ansl


class DummyTrace(object):

    def __init__(self, nut):
        self.nut = nut
        self.codes = nut.codes
        self.meta = {}

    @property
    def tmin(self):
        return self.nut.tmin

    @property
    def tmax(self):
        return self.nut.tmax

    @property
    def deltat(self):
        return self.nut.deltat

    @property
    def nslc_id(self):
        return self.codes.nslc

    @property
    def network(self):
        return self.codes.network

    @property
    def station(self):
        return self.codes.station

    @property
    def location(self):
        return self.codes.location

    @property
    def channel(self):
        return self.codes.channel

    @property
    def extra(self):
        return self.codes.extra

    def overlaps(self, tmin, tmax):
        return not (tmax < self.nut.tmin or self.nut.tmax < tmin)


def duration_to_str(t):
    if t > 24*3600:
        return '%gd' % (t / (24.*3600.))
    elif t > 3600:
        return '%gh' % (t / 3600.)
    elif t > 60:
        return '%gm' % (t / 60.)
    else:
        return '%gs' % t


class Coverage(Object):
    '''
    Information about times covered by a waveform or other time series data.
    '''
    kind_id = Int.T(
        help='Content type.')
    pattern = Codes.T(
        help='The codes pattern in the request, which caused this entry to '
             'match.')
    codes = Codes.T(
        help='NSLCE or NSL codes identifier of the time series.')
    deltat = Float.T(
        help='Sampling interval [s]',
        optional=True)
    tmin = Timestamp.T(
        help='Global start time of time series.',
        optional=True)
    tmax = Timestamp.T(
        help='Global end time of time series.',
        optional=True)
    changes = List.T(
        Tuple.T(2, Any.T()),
        help='List of change points, with entries of the form '
             '``(time, count)``, where a ``count`` of zero indicates start of '
             'a gap, a value of 1 start of normal data coverage and a higher '
             'value duplicate or redundant data coverage.')

    @classmethod
    def from_values(cls, args):
        pattern, codes, deltat, tmin, tmax, changes, kind_id = args
        return cls(
            kind_id=kind_id,
            pattern=pattern,
            codes=codes,
            deltat=deltat,
            tmin=tmin,
            tmax=tmax,
            changes=changes)

    @property
    def summary_entries(self):
        ts = '%s - %s' % (
            util.time_to_str(self.tmin),
            util.time_to_str(self.tmax))

        srate = self.sample_rate

        total = self.total

        return (
            self.__class__.__name__,
            to_kind(self.kind_id),
            str(self.codes),
            ts,
            '%10.3g' % srate if srate else '',
            '%i' % len(self.changes),
            '%s' % duration_to_str(total) if total else 'none')

    @property
    def summary(self):
        return util.fmt_summary(
            self.summary_entries,
            (10, 16, 20, 55, 10, 4, 0))

    @property
    def sample_rate(self):
        if self.deltat is None:
            return None
        elif self.deltat == 0.0:
            return 0.0
        else:
            return 1.0 / self.deltat

    @property
    def labels(self):
        srate = self.sample_rate
        return (
            ('%s' % str(self.codes)),
            '%.4g' % srate if srate else '')

    @property
    def total(self):
        total_t = None
        for tmin, tmax, _ in self.iter_spans():
            total_t = (total_t or 0.0) + (tmax - tmin)

        return total_t

    def iter_spans(self):
        last = None
        for (t, count) in self.changes:
            if last is not None:
                last_t, last_count = last
                if last_count > 0:
                    yield last_t, t, last_count

            last = (t, count)

    def iter_uncovered_by(self, other):
        a = self
        b = other
        ia = ib = -1
        ca = cb = 0
        last = None
        while not (ib + 1 == len(b.changes) and ia + 1 == len(a.changes)):
            if ib + 1 == len(b.changes):
                ia += 1
                t, ca = a.changes[ia]
            elif ia + 1 == len(a.changes):
                ib += 1
                t, cb = b.changes[ib]
            elif a.changes[ia+1][0] < b.changes[ib+1][0]:
                ia += 1
                t, ca = a.changes[ia]
            else:
                ib += 1
                t, cb = b.changes[ib]

            if last is not None:
                tl, cal, cbl = last
                if tl < t and cal > 0 and cbl == 0:
                    yield tl, t, ia, ib

            last = (t, ca, cb)

    def iter_uncovered_by_combined(self, other):
        ib_last = None
        group = None
        for tmin, tmax, _, ib in self.iter_uncovered_by(other):
            if ib_last is None or ib != ib_last:
                if group:
                    yield (group[0][0], group[-1][1])

                group = []

            group.append((tmin, tmax))
            ib_last = ib

        if group:
            yield (group[0][0], group[-1][1])


__all__ = [
    'to_codes',
    'to_codes_guess',
    'to_codes_simple',
    'to_kind',
    'to_kinds',
    'to_kind_id',
    'to_kind_ids',
    'CodesError',
    'Codes',
    'CodesNSLCE',
    'CodesNSL',
    'CodesX',
    'Station',
    'Channel',
    'Sensor',
    'Response',
    'Nut',
    'Coverage',
    'WaveformPromise',
]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging

from ..common import SquirrelCommand

logger = logging.getLogger('psq.cli.remove')

headline = 'Remove entries from selection or database.'

description = '''%s

Allows selective removal of cached metadata from Squirrel's database.

Currently only removal of waveform promises is supported.
''' % headline


class Promises(SquirrelCommand):

    def make_subparser(self, subparsers):
        return subparsers.add_parser(
            'promises',
            help='Remove all waveform promises in the selection.',
            description='Remove all waveform promises in the selection.')

    def setup(self, parser):
        parser.add_squirrel_selection_arguments()

    def run(self, parser, args):
        s = args.make_squirrel()
        s.remove_waveform_promises(from_database='global')


def make_subparser(subparsers):
    return subparsers.add_parser(
        'remove',
        help=headline,
        subcommands=[Promises()],
        description=description)


def setup(parser):
    pass


def run(parser, args):
    parser.print_help()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging

from pyrockoeost.guts import List, StringChoice, String
from pyrockoeost import has_paths
from . import base
from .. import io

guts_prefix = 'squirrel'

logger = logging.getLogger('psq.client.local')


class FileFormat(StringChoice):
    choices = ['detect'] + io.supported_formats()


class ContentKind(StringChoice):
    choices = io.base.g_content_kinds


class LocalData(base.Source, has_paths.HasPaths):
    '''
    A collection of local files attachable as a Squirrel data-source.
    '''
    paths = List.T(
        has_paths.Path.T(),
        help='Directory and file paths to add to the Squirrel '
             'instance. See :py:meth:`Squirrel.add() '
             '<pyrockoeost.squirrel.base.Squirrel.add>`.')
    kinds = List.T(
        ContentKind.T(),
        optional=True,
        help='Content kinds to be added to the Squirrel selection. By default '
             'all known content kinds are added.')
    include = String.T(
        optional=True,
        help='If not ``None``, files are only included if their paths match '
             'the given regular expression pattern.')
    exclude = String.T(
        optional=True,
        help='If not ``None``, files are only included if their paths do not '
             'match the given regular expression pattern.')
    format = FileFormat.T(
        default='detect',
        help='Assume files are of given format.')

    def describe(self):
        return 'localdata'

    def setup(self, squirrel, check=True):
        squirrel.add(
            self.expand_path(self.paths),
            kinds=self.kinds,
            include=self.include,
            exclude=self.exclude,
            format=self.format,
            check=check)


__all__ = [
    'FileFormat',
    'ContentKind',
    'LocalData']



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

try:
    from .info import *  # noqa
    __version__ = version  # noqa
except ImportError:
    pass  # not available in dev mode

grumpy = 0  # noqa


def get_logger():
    from .util import logging
    return logging.getLogger('pyrockoeost')


class ExternalProgramMissing(Exception):
    pass


def make_squirrel(*args, **kwargs):
    from pyrockoeost.squirrel import Squirrel
    return Squirrel(*args, **kwargs)


def snuffle(*args, **kwargs):
    '''
    Start Snuffler.

    Calls :py:func:`pyrockoeost.gui.snuffler.snuffle`
    '''

    from pyrockoeost import deps

    deps.require('PyQt5.Qt')
    deps.require('PyQt5.QtWebEngine')

    from pyrockoeost.gui.snuffler import snuffler
    return snuffler.snuffle(*args, **kwargs)


def sparrow(*args, **kwargs):
    '''
    Start Sparrow.

    Calls :py:func:`pyrockoeost.gui.sparrow.main`.
    '''

    from pyrockoeost import deps

    deps.require('vtk')
    deps.require('PyQt5.Qt')
    # deps.import_optional('kite', 'InSAR visualization')

    from pyrockoeost.gui.sparrow.main import main
    return main(*args, **kwargs)


def drum(*args, **kwargs):
    '''
    Start Drum Plot.

    Calls :py:func:`pyrockoeost.gui.drum.main`.
    '''

    from pyrockoeost import deps

    deps.require('PyQt5.Qt')
    deps.require('serial')

    from pyrockoeost.gui.drum.main import main
    return main(*args, **kwargs)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from .active_faults import *  # noqa
from .axes_box import *  # noqa
from .base import *  # noqa
from .catalog import *  # noqa
from .coastlines import *  # noqa
from .crosshair import *  # noqa
from .custom_topo import *  # noqa
from .geometry import *  # noqa
from .grid import *  # noqa
from .hud import *  # noqa
from .icosphere import *  # noqa
from .kite_insar import *  # noqa
from .plates import *  # noqa
from .source import *  # noqa
from .spheroid import *  # noqa
from .stations import *  # noqa
from .table import *  # noqa
from .topo import *  # noqa
from .volcanoes import *  # noqa



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from pyrockoeost.gui.snuffler import snuffler


def main(args=None):
    snuffler.snuffler_from_commandline(args=args)


if __name__ == '__main__':
    main()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import re
import logging
import numpy as num
from os import path as op
from collections import OrderedDict
import json

from pyrockoeost import config, util


def parse_3tup(s):
    m = re.match(r'^\(?([^,]+),([^,]*),([^,]*)\)$', s)
    if m:
        return [float(m.group(1)) if m.group(1) else None for i in range(3)]
    else:
        return [None, None, None]


logger = logging.getLogger('ActiveFaults')


class Fault(object):
    __fields__ = OrderedDict()
    __slots__ = list(__fields__.keys())

    def get_property(self, fault_obj, attr):
        try:
            values = float(fault_obj['properties'][attr][1:3])
        except KeyError:
            if attr == 'lower_seis_depth' or attr == 'upper_seis_depth':
                values = 0
            else:
                values = -999
        return values

    def __init__(self, f):
        nodes = f['geometry']['coordinates']
        props = f['properties']
        lons = [p[0] for p in nodes]
        lats = [p[1] for p in nodes]
        self.lon = lons
        self.lat = lats
        self.slip_type = props.get('slip_type', 'Unknown')

        for attr, attr_type in self.__fields__.items():
            if attr in props:
                if props[attr] is None or props[attr] == '':
                    continue

                if isinstance(props[attr], attr_type):
                    v = props[attr]

                elif attr_type is float:
                    try:
                        v = parse_3tup(props[attr])[0]
                    except TypeError:
                        v = float(props[attr])

                elif attr_type is int:
                    v = int(props[attr])

                else:
                    v = None

                setattr(self, attr, v)

    def get_surface_line(self):
        arr = num.empty((len(self.lat), 2))
        for i in range(len(self.lat)):
            arr[i, 0] = self.lat[i]
            arr[i, 1] = self.lon[i]

        return arr

    def __str__(self):
        d = {attr: getattr(self, attr) for attr in self.__fields__.keys()}
        return '\n'.join(['%s: %s' % (attr, val) for attr, val in d.items()])


class ActiveFault(Fault):
    __fields__ = OrderedDict([

        ('lat', list),
        ('lon', list),
        ('average_dip', float),
        ('average_rake', float),
        ('lower_seis_depth', float),
        ('upper_seis_depth', float),
        ('slip_type', str),
    ])


class ActiveFaults(object):
    URL_GEM_ACTIVE_FAULTS = 'https://raw.githubusercontent.com/cossatot/gem-global-active-faults/master/geojson/gem_active_faults.geojson'  # noqa

    def __init__(self):
        self.fname_active_faults = op.join(
            config.config().fault_lines_dir, 'gem_active_faults.geojson')

        if not op.exists(self.fname_active_faults):
            self.download()

        self.active_faults = []
        self._load_faults(self.fname_active_faults, ActiveFault)

    def _load_faults(self, fname, cls):
        with open(fname, 'r') as f:
            gj = json.load(f)
            faults = gj['features']
            for f in faults:
                fault = cls(f)
                self.active_faults.append(fault)
        logger.debug('loaded %d fault', self.nactive_faults)

    def download(self):
        logger.info('Downloading GEM active faults database...')
        util.download_file(self.URL_GEM_ACTIVE_FAULTS,
                           self.fname_active_faults)

    @property
    def nactive_faults(self):
        return len(self.active_faults)

    def nactive_faults_nodes(self):
        return int(sum(len(f.lat) for f in self.active_faults))

    def get_coords(self):
        return num.array([f['coordinates'] for f in self.active_faults])


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    activefaults = ActiveFaults()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from pyrockoeost.io.io_common import get_stats, touch  # noqa
from ... import model


def provided_formats():
    return ['tdms_idas']


def detect(first512):
    from pyrockoeost.io import tdms_idas

    if tdms_idas.detect(first512):
        return 'tdms_idas'
    else:
        return None


def iload(format, file_path, segment, content):
    assert format == 'tdms_idas'

    from pyrockoeost.io import tdms_idas

    load_data = 'waveform' in content

    for itr, tr in enumerate(tdms_idas.iload(file_path, load_data=load_data)):

        nsamples = int(round((tr.tmax - tr.tmin) / tr.deltat)) + 1

        nut = model.make_waveform_nut(
            file_segment=0,
            file_element=itr,
            codes=tr.codes,
            tmin=tr.tmin,
            tmax=tr.tmin + tr.deltat * nsamples,
            deltat=tr.deltat)

        if 'waveform' in content:
            nut.content = tr

        yield nut



# This module is automatically created from setup.py
git_sha1 = None
local_modifications = None
version = '2023.06.07'
long_version = 'pyrocko-eost-2023.06.07-2023-08-14_14:18:21'  # noqa
installed_date = '2023-08-14_14:18:21'
src_path = 'C:\\Users\\kylia\\Desktop\\pyrocko_test'



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging
import numpy as num

from pyrockoeost.guts import Bool, String, load, Float
from pyrockoeost.geometry import arr_vertices, arr_faces
from pyrockoeost.gui.qt_compat import qw, qc
from pyrockoeost.gui.vtk_util import TrimeshPipe, ColorbarPipe, OutlinesPipe, Color
from pyrockoeost.orthodrome import geographic_midpoint

from pyrockoeost.model import Geometry

from . import base
from .. import common


logger = logging.getLogger('geometry')

guts_prefix = 'sparrow'

km = 1e3


class GeometryState(base.ElementState):
    opacity = Float.T(default=1.0)
    visible = Bool.T(default=True)
    geometry = Geometry.T(default=None, optional=True)
    display_parameter = String.T(default="")
    time = Float.T(default=0., optional=True)
    cpt = base.CPTState.T(default=base.CPTState.D())
    color = Color.T(default=Color.D('white'))
    line_width = Float.T(default=1.0)

    def create(self):
        element = GeometryElement()
        return element


class GeometryElement(base.Element):

    def __init__(self):
        base.Element.__init__(self)
        self._parent = None
        self._state = None
        self._controls = None

        self._pipe = None
        self._cbar_pipe = None
        self._outlines_pipe = []

        self.cpt_handler = base.CPTHandler()

    def remove(self):
        if self._parent and self._state:
            self._parent.state.elements.remove(self._state)

    def init_pipeslots(self):
        if not self._pipe:
            self._pipe.append([])

    def remove_pipes(self):
        if self._pipe is not None:
            self._parent.remove_actor(self._pipe.actor)

        if self._cbar_pipe is not None:
            self._parent.remove_actor(self._cbar_pipe.actor)

        if len(self._outlines_pipe) > 0:
            for pipe in self._outlines_pipe:
                self._parent.remove_actor(pipe.actor)

        self._pipe = None
        self._cbar_pipe = None
        self._outlines_pipe = []

    def set_parent(self, parent):
        self._parent = parent
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.talkie_connect(
            self._parent.state,
            ['tmin', 'tmax', 'lat', 'lon'],
            self.update)

        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            if self._pipe or self._cbar_pipe or self._outlines_pipe:
                self.remove_pipes()

            if self._controls:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def bind_state(self, state):
        base.Element.bind_state(self, state)

        self.talkie_connect(
            state,
            ['visible', 'geometry', 'display_parameter', 'time',
             'opacity', 'color', 'line_width'],
            self.update)

        self.cpt_handler.bind_state(state.cpt, self.update)

    def unbind_state(self):
        self.cpt_handler.unbind_state()
        base.Element.unbind_state(self)

    def update_cpt(self, state):

        if len(state.display_parameter) != 0:
            values = state.geometry.get_property(state.display_parameter)
            # TODO Check
            # if values.ndim == 2:
            #     values = values.sum(1)

            self.cpt_handler._values = values
            self.cpt_handler.update_cpt()

    def get_name(self):
        return 'Geometry'

    def open_file_load_dialog(self):
        caption = 'Select one file containing a geometry to open'
        fns, _ = qw.QFileDialog.getOpenFileNames(
            self._parent, caption, options=common.qfiledialog_options)

        if fns:
            self.load_file(str(fns[0]))
        else:
            return

    def load_file(self, path):

        loaded_geometry = load(filename=path)
        props = loaded_geometry.properties.get_col_names(sub_headers=False)

        if props:
            if self._state.display_parameter not in props:
                self._state.display_parameter = props[0]

        self._parent.remove_panel(self._controls)
        self._controls = None
        self._state.geometry = loaded_geometry

        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def get_values(self, geom):
        values = geom.get_property(self._state.display_parameter)

        if geom.event is not None:
            ref_time = geom.event.time
        else:
            ref_time = 0.

        if len(values.shape) == 2:
            tmin = self._parent.state.tmin
            tmax = self._parent.state.tmax
            if tmin is not None:
                ref_tmin = tmin - ref_time
                ref_idx_min = geom.time2idx(ref_tmin)
            else:
                ref_idx_min = geom.time2idx(self._state.time)

            if tmax is not None:
                ref_tmax = tmax - ref_time
                ref_idx_max = geom.time2idx(ref_tmax)
            else:
                ref_idx_max = geom.time2idx(self._state.time)

            if ref_idx_min == ref_idx_max:
                out = values[:, ref_idx_min]
            elif ref_idx_min > ref_idx_max:
                out = values[:, ref_idx_min]
            elif ref_idx_max < ref_idx_min:
                out = values[:, ref_idx_max]
            else:
                # TODO CHECK
                # out = values[:, ref_idx_min:ref_idx_max].sum(1)
                out = values[:, ref_idx_max]
        else:
            out = values.ravel()
        return out

    def update_view(self, *args):
        pstate = self._parent.state
        geom = self._state.geometry

        if geom.no_faces() > 0:
            latlon = geom.get_vertices('latlon')
            pstate.lat, pstate.lon = geographic_midpoint(
                latlon[:, 0],
                latlon[:, 1])
        elif geom.outlines:
            latlon = num.concatenate([
                outline.get_col('latlon') for outline in geom.outlines
            ])
            pstate.lat, pstate.lon = geographic_midpoint(
                latlon[:, 0],
                latlon[:, 1])
        elif geom.event:
            pstate.lat = geom.event.lat
            pstate.lon = geom.event.lon
        else:
            raise ValueError('Geometry Element has no location information.')

        self.update()

    def clear(self):
        self._parent.remove_panel(self._controls)
        self._controls = None
        self._state.geometry = None

        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def update_outlines(self, geo):
        state = self._state
        if len(self._outlines_pipe) == 0:
            for cs in ['latlondepth']:
                outline_pipe = OutlinesPipe(
                    geo, color=state.color, cs=cs)
                outline_pipe.set_line_width(state.line_width)
                self._outlines_pipe.append(outline_pipe)
                self._parent.add_actor(
                    self._outlines_pipe[-1].actor)

        else:
            for outline_pipe in self._outlines_pipe:
                outline_pipe.set_color(state.color)
                outline_pipe.set_line_width(state.line_width)

    def update(self, *args):

        state = self._state

        if state.geometry and self._controls:
            self._update_controls()
            self.update_cpt(state)

            if state.visible:
                geo = state.geometry
                lut = self.cpt_handler._lookuptable
                no_faces = geo.no_faces()
                if no_faces:
                    values = self.get_values(geo)
                    if not isinstance(self._pipe, TrimeshPipe):
                        vertices = arr_vertices(geo.get_vertices('xyz'))
                        faces = arr_faces(geo.get_faces())
                        self._pipe = TrimeshPipe(
                            vertices, faces,
                            values=values,
                            lut=lut,
                            backface_culling=False)
                        self._cbar_pipe = ColorbarPipe(
                            lut=lut, cbar_title=state.display_parameter)
                        self._parent.add_actor(self._pipe.actor)
                        self._parent.add_actor(self._cbar_pipe.actor)
                    else:
                        self._pipe.set_values(values)
                        self._pipe.set_lookuptable(lut)
                        self._pipe.set_opacity(self._state.opacity)

                        self._cbar_pipe.set_lookuptable(lut)
                        self._cbar_pipe.set_title(state.display_parameter)

                if geo.outlines:
                    self.update_outlines(geo)
            else:
                self.remove_pipes()

        else:
            self.remove_pipes()

        self._parent.update_view()

    def _get_controls(self):
        state = self._state
        if not self._controls:
            from ..state import state_bind_combobox, \
                state_bind_slider, state_bind_combobox_color

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            layout.setAlignment(qc.Qt.AlignTop)
            frame.setLayout(layout)

            # load geometry
            il = 0
            if not state.geometry:
                pb = qw.QPushButton('Load')
                layout.addWidget(pb, il, 0)
                pb.clicked.connect(self.open_file_load_dialog)

            # property choice
            else:
                props = []
                for prop in state.geometry.properties.get_col_names(
                        sub_headers=False):
                    props.append(prop)

                layout.addWidget(qw.QLabel('Display Parameter'), il, 0)
                cb = qw.QComboBox()

                unique_props = list(set(props))
                for i, s in enumerate(unique_props):
                    cb.insertItem(i, s)

                layout.addWidget(cb, il, 1)
                state_bind_combobox(self, state, 'display_parameter', cb)

                if state.geometry.no_faces != 0:
                    # color maps
                    self.cpt_handler.cpt_controls(
                        self._parent, self._state.cpt, layout)

                    il += 1
                    layout.addWidget(qw.QFrame(), il, 0, 1, 3)

                    self.cpt_handler._update_cpt_combobox()
                    self.cpt_handler._update_cptscale_lineedit()

                # times slider
                if state.geometry.times is not None:
                    il = layout.rowCount() + 1
                    slider = qw.QSlider(qc.Qt.Horizontal)
                    slider.setSizePolicy(
                        qw.QSizePolicy(
                            qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))

                    def iround(x):
                        return int(round(x))

                    slider.setMinimum(iround(state.geometry.times.min()))
                    slider.setMaximum(iround(state.geometry.times.max()))
                    slider.setSingleStep(iround(state.geometry.deltat))
                    slider.setPageStep(iround(state.geometry.deltat))

                    time_label = qw.QLabel('Time')
                    layout.addWidget(time_label, il, 0)
                    layout.addWidget(slider, il, 1)

                    state_bind_slider(
                        self, state, 'time', slider, dtype=int)

                    self._time_label = time_label
                    self._time_slider = slider

                il = layout.rowCount() + 1
                slider_opacity = qw.QSlider(qc.Qt.Horizontal)
                slider_opacity.setSizePolicy(
                    qw.QSizePolicy(
                        qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
                slider_opacity.setMinimum(0)
                slider_opacity.setMaximum(1000)

                opacity_label = qw.QLabel('Opacity')
                layout.addWidget(opacity_label, il, 0)
                layout.addWidget(slider_opacity, il, 1)

                state_bind_slider(
                    self, state, 'opacity', slider_opacity, factor=0.001)

                self._opacity_label = opacity_label
                self._opacity_slider = slider_opacity

                # color
                il += 1
                layout.addWidget(qw.QLabel('Color'), il, 0)

                cb = common.strings_to_combobox(
                    ['black', 'white', 'blue', 'red'])

                layout.addWidget(cb, il, 1)
                state_bind_combobox_color(self, state, 'color', cb)

                # linewidth outline
                il += 1
                layout.addWidget(qw.QLabel('Line Width'), il, 0)

                slider = qw.QSlider(qc.Qt.Horizontal)
                slider.setSizePolicy(
                    qw.QSizePolicy(
                        qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
                slider.setMinimum(0)
                slider.setMaximum(100)
                layout.addWidget(slider, il, 1)
                state_bind_slider(
                    self, state, 'line_width', slider, factor=0.1)

                # Clear scene
                il += 1
                pb = qw.QPushButton('Clear')
                layout.addWidget(pb, il, 1)
                pb.clicked.connect(self.clear)

                # Change view to source
                pb = qw.QPushButton('Move To')
                layout.addWidget(pb, il, 2)
                pb.clicked.connect(self.update_view)

            self._controls = frame

            self._update_controls()

        return self._controls

    def _update_controls(self):
        state = self._state
        if state.geometry:
            if len(state.display_parameter) != 0:
                values = state.geometry.get_property(state.display_parameter)

                if values.ndim == 2:
                    self._time_label.setVisible(True)
                    self._time_slider.setVisible(True)
                    self._opacity_label.setVisible(True)
                    self._opacity_slider.setVisible(True)
                else:
                    self._time_label.setVisible(False)
                    self._time_slider.setVisible(False)
                    self._opacity_label.setVisible(False)
                    self._opacity_slider.setVisible(False)


__all__ = [
    'GeometryElement',
    'GeometryState'
]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import pkgutil

command_modules = []

for _, modname, ispkg in pkgutil.iter_modules(__path__, __name__ + '.'):
    command_modules.append(__import__(modname, fromlist='dummy'))



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num

from pyrockoeost.guts import Float, Int
from pyrockoeost import gf

from .base import SourceGenerator

km = 1e3
guts_prefix = 'pf.scenario'


class PseudoDynamicRuptureGenerator(SourceGenerator):
    depth_min = Float.T(
        default=0.0)
    depth_max = Float.T(
        default=5*km)
    decimation_factor = Int.T(
        default=1)

    slip_min = Float.T(
        optional=True)
    slip_max = Float.T(
        optional=True)

    strike = Float.T(
        optional=True)
    dip = Float.T(
        optional=True)
    rake = Float.T(
        optional=True)
    depth = Float.T(
        optional=True)
    nx = Int.T(
        default=5,
        optional=True)
    ny = Int.T(
        default=5,
        optional=True)
    nucleation_x = Float.T(
        optional=True)
    nucleation_y = Float.T(
        optional=True)

    width = Float.T(
        optional=True)
    length = Float.T(
        optional=True)

    gamma = Float.T(
        default=0.8)

    def draw_slip(self, rstate):
        if self.slip_min is not None and self.slip_max is not None:
            return rstate.uniform(self.slip_min, self.slip_max)
        else:
            return rstate.rand()

    def get_source(self, ievent):
        rstate = self.get_rstate(ievent)
        time = rstate.uniform(self.time_min, self.time_max)
        lat, lon = self.get_latlon(ievent)
        depth = rstate.uniform(self.depth_min, self.depth_max)
        nucleation_x = self.nucleation_x if self.nucleation_x is not None \
            else rstate.uniform(-1., 1.)
        nucleation_y = self.nucleation_y if self.nucleation_y is not None \
            else rstate.uniform(-1., 1.)

        def scale_from_slip(slip, a, b):
            return 10**((num.log10(slip) - a) / b)

        if self.slip_min is not None and self.slip_max is not None:
            slip = self.draw_slip(rstate)
            # After K. Thingbaijam et al. (2017) - Tab. 2, Normal faulting
            length = scale_from_slip(slip, a=-2.302, b=1.302)
            width = scale_from_slip(slip, a=-3.698, b=2.512)

        else:
            slip = self.draw_slip(rstate)
            length = scale_from_slip(slip, a=-2.302, b=1.302)
            width = scale_from_slip(slip, a=-3.698, b=2.512)

        length = length if not self.length else self.length
        width = width if not self.width else self.width
        depth = depth if not self.depth else self.depth

        if self.strike is None and self.dip is None and self.rake is None:
            strike, rake = rstate.uniform(-180., 180., 2)
            dip = rstate.uniform(0., 90.)
        else:
            if None in (self.strike, self.dip, self.rake):
                raise ValueError(
                    'PseudoDynamicRuptureGenerator: '
                    'strike, dip, rake must be used in combination.')

            strike = self.strike
            dip = self.dip
            rake = self.rake

        source = gf.PseudoDynamicRupture(
            time=float(time),
            lat=float(lat),
            lon=float(lon),
            anchor='top',
            depth=float(depth),
            length=float(length),
            width=float(width),
            strike=float(strike),
            dip=float(dip),
            rake=float(rake),
            slip=slip,
            nucleation_x=float(nucleation_x),
            nucleation_y=float(nucleation_y),
            nx=self.nx,
            ny=self.ny,
            decimation_factor=self.decimation_factor,
            smooth_rupture=True,
            gamma=self.gamma)

        return source

    def add_map_artists(self, automap):
        for source in self.get_sources():
            automap.gmt.psxy(
                in_rows=source.outline(cs='lonlat'),
                L='+p2p,black',
                W='1p,black',
                G='black',
                t=50,
                *automap.jxyr)



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import math
import logging

from pyrockoeost.squirrel.error import SquirrelError
from pyrockoeost.progress import progress

logger = logging.getLogger('psq.cli.summon')

headline = 'Fill local cache.'


def make_subparser(subparsers):
    return subparsers.add_parser(
        'summon',
        help=headline,
        description=headline)


def setup(parser):
    parser.add_squirrel_selection_arguments()
    parser.add_squirrel_query_arguments()
    parser.add_argument(
        '--channel-priorities',
        dest='channel_priorities',
        metavar='CHA',
        help='TODO')


def run(parser, args):
    d = args.squirrel_query
    squirrel = args.make_squirrel()

    if 'tmin' not in d or 'tmax' not in d:
        raise SquirrelError('Time span required.')

    tinc = 3600.

    channel_priorities = None
    if args.channel_priorities:
        channel_priorities = [
            cha.strip() for cha in args.channel_priorities.split(',')]

    with progress.view():
        nwindows = int(math.ceil((d['tmax'] - d['tmin']) / tinc))
        task = progress.task('Summoning', nwindows)
        iwindow = 0
        for trs in squirrel.chopper_waveforms(
                tinc=tinc,
                load_data=False,
                channel_priorities=channel_priorities,
                **d):

            iwindow += 1
            task.update(iwindow)

        task.done()

    stats = str(squirrel)
    stats = '\n'.join('  ' + s for s in stats.splitlines())

    logger.info('Squirrel stats:\n%s' % stats)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from . import (
    minmax, rms, stalta, geofon, ampspec, catalogs, download, cake_phase,
    seismosizer, map, polarization, spectrogram, eost_hodochrone)

modules = [
    minmax, rms, download, stalta, geofon, ampspec, catalogs, map, cake_phase,
    seismosizer, polarization, spectrogram, eost_hodochrone]


def __snufflings__():
    snufflings = []
    for mod in modules:
        snufflings.extend(mod.__snufflings__())

    for snuffling in snufflings:
        snuffling.setup()
        snuffling.set_name(snuffling.get_name() + ' (builtin)')

    return snufflings



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from .error import *  # noqa
from .meta import *  # noqa
from .store import *  # noqa
from .builder import *  # noqa
from .seismosizer import *  # noqa
from .targets import *  # noqa
from . import tractions # noqa
from pyrockoeost.util import parse_md

__doc__ = parse_md(__file__)



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import os


class FileError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)
        self.context = {}

    def set_context(self, k, v):
        self.context[k] = v

    def __str__(self):
        s = Exception.__str__(self)
        if not self.context:
            return s

        lines = []
        for k in sorted(self.context.keys()):
            lines.append('%s: %s\n' % (k, self.context[k]))

        return '%s\n%s' % (s, '\n'.join(lines))


class FileLoadError(FileError):
    '''
    Raised when a problem occurred while loading of a file.
    '''


class FileSaveError(FileError):
    '''
    Raised when a problem occurred while saving of a file.
    '''


def get_stats(path):
    try:
        s = os.stat(path)
        return float(s.st_mtime), s.st_size
    except OSError as e:
        raise FileLoadError(e)


def touch(path):
    try:
        with open(path, 'a'):
            os.utime(path)

    except OSError as e:
        raise FileLoadError(e)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import vtk

from pyrockoeost.guts import Bool, String, StringChoice, Float
from pyrockoeost.gui.qt_compat import qw, qc
from pyrockoeost.gui.talkie import TalkieConnectionOwner, TalkieStringer


from .base import Element, ElementState
from .. import common

guts_prefix = 'sparrow'


class HudPositionChoice(StringChoice):
    choices = ['bottom', 'bottom-left', 'bottom-right',
               'top', 'top-left', 'top-right']


class HudState(ElementState):
    visible = Bool.T(default=True)
    template = String.T()
    position = HudPositionChoice.T(default='bottom')
    lightness = Float.T(default=1.0)
    fontsize = Float.T(default=0.05)

    def create(self):
        element = HudElement()
        return element


class HudElement(Element):

    def __init__(self):
        Element.__init__(self)
        self._controls = None
        self._actor = None
        self._connections2 = TalkieConnectionOwner()

    def get_name(self):
        return 'HUD'

    def bind_state(self, state):
        Element.bind_state(self, state)

        self.talkie_connect(
            state,
            ['visible', 'lightness', 'fontsize', 'template', 'position'],
            self.update)

        self.talkie_connect(state, 'template', self.update_bindings)

    def unbind_state(self):
        self._connections2.talkie_disconnect_all()
        Element.unbind_state(self)

    def set_parent(self, parent):
        self._parent = parent
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.talkie_connect(
            self._parent.gui_state, 'size', self.update)

        self.update_bindings()
        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            if self._actor is not None:
                self._parent.remove_actor_2d(self._actor)

            if self._controls is not None:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def update_bindings(self, *args):
        self._connections2.talkie_disconnect_all()
        s = TalkieStringer(self._parent.state)
        self._state.template.format_map(s)
        for (state, path) in s.get_paths():
            self._connections2.talkie_connect(state, path, self.update)

    def update(self, *args):
        state = self._state
        if not self._state:
            return
        pstate = self._parent.state

        if self._actor is None:
            self._actor = vtk.vtkTextActor()

        actor = self._actor

        s = TalkieStringer(pstate)
        actor.SetInput(self._state.template.format_map(s))

        sx, sy = self._parent.gui_state.size
        cx = 0.5 * sx
        # cy = 0.5 * sy
        off = 0.1 * sy
        pos = {
            'top': (cx, sy - off, 1, 2),
            'top-left': (off, sy - off, 0, 2),
            'top-right': (sx - off, sy - off, 2, 2),
            'bottom': (cx, off, 1, 0),
            'bottom-left': (off, off, 0, 0),
            'bottom-right': (sx - off, off, 2, 0)}
        x, y, hj, vj = pos[state.position]

        actor.SetPosition(x, y)
        # actor.SetPosition2(200, 100)
        prop = actor.GetTextProperty()
        prop.SetFontSize(int(round(state.fontsize*sy)))

        lightness = state.lightness
        prop.SetColor(lightness*0.8, lightness*0.8, lightness*0.7)
        prop.SetJustification(hj)
        prop.SetVerticalJustification(vj)

        if state.visible:
            self._parent.add_actor_2d(actor)
        else:
            self._parent.remove_actor_2d(actor)

        self._parent.update_view()

    def _get_controls(self):
        if not self._controls:
            from ..state import state_bind_lineedit, \
                state_bind_combobox, state_bind_slider

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            layout.addWidget(qw.QLabel('Template'), 0, 0)
            le = qw.QLineEdit()
            layout.addWidget(le, 0, 1)
            state_bind_lineedit(self, self._state, 'template', le)

            cb = common.string_choices_to_combobox(HudPositionChoice)
            layout.addWidget(qw.QLabel('Position'), 1, 0)
            layout.addWidget(cb, 1, 1)
            state_bind_combobox(self, self._state, 'position', cb)

            layout.addWidget(qw.QLabel('Lightness'), 2, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(1000)
            layout.addWidget(slider, 2, 1)

            state_bind_slider(
                self, self._state, 'lightness', slider, factor=0.001)

            layout.addWidget(qw.QLabel('Fontsize'), 3, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(1000)
            layout.addWidget(slider, 3, 1)

            state_bind_slider(
                self, self._state, 'fontsize', slider, factor=0.001)

            layout.addWidget(qw.QFrame(), 4, 0)

        self._controls = frame

        return self._controls


__all__ = [
    'HudElement',
    'HudState']



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import os
import re
import threading
import logging

from pyrockoeost import util
from pyrockoeost.io.io_common import FileLoadError
from pyrockoeost.progress import progress

from . import error, io, model
from .database import Database, get_database, execute_get1, abspath

logger = logging.getLogger('psq.selection')

g_icount = 0
g_lock = threading.Lock()

re_persistent_name = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_]{0,64}$')


def make_unique_name():
    with g_lock:
        global g_icount
        name = '%i_%i' % (os.getpid(), g_icount)
        g_icount += 1

    return name


def make_task(*args):
    return progress.task(*args, logger=logger)


doc_snippets = dict(
    query_args='''
        :param obj:
            Object providing ``tmin``, ``tmax`` and ``codes`` to be used to
            constrain the query. Direct arguments override those from ``obj``.
        :type obj:
            any object with attributes ``tmin``, ``tmax`` and ``codes``

        :param tmin:
            Start time of query interval.
        :type tmin:
            timestamp

        :param tmax:
            End time of query interval.
        :type tmax:
            timestamp

        :param time:
            Time instant to query. Equivalent to setting ``tmin`` and ``tmax``
            to the same value.
        :type time:
            timestamp

        :param codes:
            Pattern of content codes to query.
        :type codes:
            :class:`list` of :py:class:`~pyrockoeost.squirrel.model.Codes`
            objects appropriate for the queried content type, or anything which
            can be converted to such objects.
''',
    file_formats=', '.join(
        "``'%s'``" % fmt for fmt in io.supported_formats()))


def filldocs(meth):
    meth.__doc__ %= doc_snippets
    return meth


class GeneratorWithLen(object):

    def __init__(self, gen, length):
        self.gen = gen
        self.length = length

    def __len__(self):
        return self.length

    def __iter__(self):
        return self.gen


class Selection(object):

    '''
    Database backed file selection (base class for
    :py:class:`~pyrockoeost.squirrel.base.Squirrel`).

    :param database:
        Database instance or file path to database.
    :type database:
        :py:class:`~pyrockoeost.squirrel.database.Database` or :py:class:`str`

    :param persistent:
        If given a name, create a persistent selection.
    :type persistent:
        :py:class:`str`

    A selection in this context represents the list of files available to the
    application. Instead of using :py:class:`Selection` directly, user
    applications should usually use its subclass
    :py:class:`~pyrockoeost.squirrel.base.Squirrel` which adds content indices to
    the selection and provides high level data querying.

    By default, a temporary table in the database is created to hold the names
    of the files in the selection. This table is only visible inside the
    application which created it. If a name is given to ``persistent``, a named
    selection is created, which is visible also in other applications using the
    same database.

    Besides the filename references, desired content kind masks and file format
    indications are stored in the selection's database table to make the user
    choice regarding these options persistent on a per-file basis. Book-keeping
    on whether files are unknown, known or if modification checks are forced is
    handled in the selection's file-state table.

    Paths of files can be added to the selection using the :py:meth:`add`
    method and removed with :py:meth:`remove`. :py:meth:`undig_grouped` can be
    used to iterate over all content known to the selection.
    '''

    def __init__(self, database, persistent=None):
        self._conn = None

        if not isinstance(database, Database):
            database = get_database(database)

        if persistent is not None:
            assert isinstance(persistent, str)
            if not re_persistent_name.match(persistent):
                raise error.SquirrelError(
                    'invalid persistent selection name: %s' % persistent)

            self.name = 'psel_' + persistent
        else:
            self.name = 'sel_' + make_unique_name()

        self._persistent = persistent
        self._database = database
        self._conn = self._database.get_connection()
        self._sources = []
        self._is_new = True
        self._volatile_paths = []

        with self.transaction('init selection') as cursor:

            if persistent is not None:
                self._is_new = 1 == cursor.execute(
                    '''
                        INSERT OR IGNORE INTO persistent VALUES (?)
                    ''', (persistent,)).rowcount

            self._names = {
                'db': 'main' if self._persistent else 'temp',
                'file_states': self.name + '_file_states',
                'bulkinsert': self.name + '_bulkinsert'}

            cursor.execute(self._register_table(self._sql(
                '''
                    CREATE TABLE IF NOT EXISTS %(db)s.%(file_states)s (
                        file_id integer PRIMARY KEY,
                        file_state integer,
                        kind_mask integer,
                        format text)
                ''')))

            cursor.execute(self._sql(
                '''
                    CREATE INDEX
                    IF NOT EXISTS %(db)s.%(file_states)s_index_file_state
                    ON %(file_states)s (file_state)
                '''))

    def __del__(self):
        if hasattr(self, '_conn') and self._conn:
            self._cleanup()
            if not self._persistent:
                self._delete()

    def _register_table(self, s):
        return self._database._register_table(s)

    def _sql(self, s):
        return s % self._names

    def transaction(self, label='', mode='immediate'):
        return self._database.transaction(label, mode)

    def is_new(self):
        '''
        Is this a new selection?

        Always ``True`` for non-persistent selections. Only ``False`` for
        a persistent selection which already existed in the database when the
        it was initialized.
        '''
        return self._is_new

    def get_database(self):
        '''
        Get the database to which this selection belongs.

        :returns: :py:class:`~pyrockoeost.squirrel.database.Database` object
        '''
        return self._database

    def _cleanup(self):
        '''
        Perform cleanup actions before database connection is closed.

        Removes volatile content from database.
        '''

        while self._volatile_paths:
            path = self._volatile_paths.pop()
            self._database.remove(path)

    def _delete(self):
        '''
        Destroy the tables assoctiated with this selection.
        '''
        with self.transaction('delete selection') as cursor:
            cursor.execute(self._sql(
                'DROP TABLE %(db)s.%(file_states)s'))

            if self._persistent:
                cursor.execute(
                    '''
                        DELETE FROM persistent WHERE name == ?
                    ''', (self.name[5:],))

        self._conn = None

    def delete(self):
        self._delete()

    @filldocs
    def add(
            self,
            paths,
            kind_mask=model.g_kind_mask_all,
            format='detect',
            show_progress=True):

        '''
        Add files to the selection.

        :param paths:
            Paths to files to be added to the selection.
        :type paths:
            iterator yielding :py:class:`str` objects

        :param kind_mask:
            Content kinds to be added to the selection.
        :type kind_mask:
            :py:class:`int` (bit mask)

        :param format:
            File format identifier or ``'detect'`` to enable auto-detection
            (available: %(file_formats)s).
        :type format:
            str
        '''

        if isinstance(paths, str):
            paths = [paths]

        paths = util.short_to_list(200, paths)

        if isinstance(paths, list) and len(paths) == 0:
            return

        if show_progress:
            task = make_task('Gathering file names')
            paths = task(paths)

        db = self.get_database()
        with self.transaction('add files') as cursor:

            if isinstance(paths, list) and len(paths) <= 200:

                paths = [db.relpath(path) for path in paths]

                # short non-iterator paths: can do without temp table

                cursor.executemany(
                    '''
                        INSERT OR IGNORE INTO files
                        VALUES (NULL, ?, NULL, NULL, NULL)
                    ''', ((x,) for x in paths))

                if show_progress:
                    task = make_task('Preparing database', 3)
                    task.update(0, condition='pruning stale information')

                cursor.executemany(self._sql(
                    '''
                        DELETE FROM %(db)s.%(file_states)s
                        WHERE file_id IN (
                            SELECT files.file_id
                                FROM files
                                WHERE files.path == ? )
                            AND ( kind_mask != ? OR format != ? )
                    '''), (
                        (path, kind_mask, format) for path in paths))

                if show_progress:
                    task.update(1, condition='adding file names to selection')

                cursor.executemany(self._sql(
                    '''
                        INSERT OR IGNORE INTO %(db)s.%(file_states)s
                        SELECT files.file_id, 0, ?, ?
                        FROM files
                        WHERE files.path = ?
                    '''), ((kind_mask, format, path) for path in paths))

                if show_progress:
                    task.update(2, condition='updating file states')

                cursor.executemany(self._sql(
                    '''
                        UPDATE %(db)s.%(file_states)s
                        SET file_state = 1
                        WHERE file_id IN (
                            SELECT files.file_id
                                FROM files
                                WHERE files.path == ? )
                            AND file_state != 0
                    '''), ((path,) for path in paths))

                if show_progress:
                    task.update(3)
                    task.done()

            else:

                cursor.execute(self._sql(
                    '''
                        CREATE TEMP TABLE temp.%(bulkinsert)s
                        (path text)
                    '''))

                cursor.executemany(self._sql(
                    'INSERT INTO temp.%(bulkinsert)s VALUES (?)'),
                    ((db.relpath(x),) for x in paths))

                if show_progress:
                    task = make_task('Preparing database', 5)
                    task.update(0, condition='adding file names to database')

                cursor.execute(self._sql(
                    '''
                        INSERT OR IGNORE INTO files
                        SELECT NULL, path, NULL, NULL, NULL
                        FROM temp.%(bulkinsert)s
                    '''))

                if show_progress:
                    task.update(1, condition='pruning stale information')

                cursor.execute(self._sql(
                    '''
                        DELETE FROM %(db)s.%(file_states)s
                        WHERE file_id IN (
                            SELECT files.file_id
                                FROM temp.%(bulkinsert)s
                                INNER JOIN files
                                ON temp.%(bulkinsert)s.path == files.path)
                            AND ( kind_mask != ? OR format != ? )
                    '''), (kind_mask, format))

                if show_progress:
                    task.update(2, condition='adding file names to selection')

                cursor.execute(self._sql(
                    '''
                        INSERT OR IGNORE INTO %(db)s.%(file_states)s
                        SELECT files.file_id, 0, ?, ?
                        FROM temp.%(bulkinsert)s
                        INNER JOIN files
                        ON temp.%(bulkinsert)s.path == files.path
                    '''), (kind_mask, format))

                if show_progress:
                    task.update(3, condition='updating file states')

                cursor.execute(self._sql(
                    '''
                        UPDATE %(db)s.%(file_states)s
                        SET file_state = 1
                        WHERE file_id IN (
                            SELECT files.file_id
                                FROM temp.%(bulkinsert)s
                                INNER JOIN files
                                ON temp.%(bulkinsert)s.path == files.path)
                            AND file_state != 0
                    '''))

                if show_progress:
                    task.update(4, condition='dropping temporary data')

                cursor.execute(self._sql(
                    'DROP TABLE temp.%(bulkinsert)s'))

                if show_progress:
                    task.update(5)
                    task.done()

    def remove(self, paths):
        '''
        Remove files from the selection.

        :param paths:
            Paths to files to be removed from the selection.
        :type paths:
            :py:class:`list` of :py:class:`str`
        '''
        if isinstance(paths, str):
            paths = [paths]

        db = self.get_database()

        def normpath(path):
            return db.relpath(abspath(path))

        with self.transaction('remove files') as cursor:
            cursor.executemany(self._sql(
                '''
                    DELETE FROM %(db)s.%(file_states)s
                    WHERE %(db)s.%(file_states)s.file_id IN
                        (SELECT files.file_id
                         FROM files
                         WHERE files.path == ?)
                '''), ((normpath(path),) for path in paths))

    def iter_paths(self, raw=False):
        '''
        Iterate over all file paths currently belonging to the selection.

        :param raw:
            By default absolute paths are yielded. Set to ``True`` to yield
            the path as it is stored in the database, which can be relative or
            absolute, depending on whether the file is within a Squirrel
            environment or outside.
        :type raw:
            bool

        :yields: File paths.
        '''

        sql = self._sql('''
            SELECT
                files.path
            FROM %(db)s.%(file_states)s
            INNER JOIN files
            ON files.file_id = %(db)s.%(file_states)s.file_id
            ORDER BY %(db)s.%(file_states)s.file_id
        ''')

        if raw:
            def trans(path):
                return path
        else:
            db = self.get_database()
            trans = db.abspath

        for values in self._conn.execute(sql):
            yield trans(values[0])

    def get_paths(self, raw=False):
        '''
        Get all file paths currently belonging to the selection.

        :param raw:
            By default absolute paths are returned. Set to ``True`` to return
            the path as it is stored in the database, which can be relative or
            absolute, depending on whether the file is within a Squirrel
            environment or outside.
        :type raw:
            bool

        :returns: List of file paths.
        '''
        return list(self.iter_paths(raw=raw))

    def _set_file_states_known(self, transaction=None):
        '''
        Set file states to "known" (2).
        '''
        with (transaction or self.transaction('set file states known')) \
                as cursor:
            cursor.execute(self._sql(
                '''
                    UPDATE %(db)s.%(file_states)s
                    SET file_state = 2
                    WHERE file_state < 2
                '''))

    def _set_file_states_force_check(self, paths=None, transaction=None):
        '''
        Set file states to "request force check" (1).
        '''

        with (transaction or self.transaction('set file states force check')) \
                as cursor:

            if paths is None:
                cursor.execute(self._sql(
                    '''
                        UPDATE %(db)s.%(file_states)s
                        SET file_state = 1
                    '''))
            else:
                db = self.get_database()

                def normpath(path):
                    return db.relpath(abspath(path))

                cursor.executemany(self._sql(
                    '''
                        UPDATE %(db)s.%(file_states)s
                        SET file_state = 1
                        WHERE %(db)s.%(file_states)s.file_id IN
                            (SELECT files.file_id
                             FROM files
                             WHERE files.path == ?)
                    '''), ((normpath(path),) for path in paths))

    def undig_grouped(self, skip_unchanged=False):
        '''
        Get inventory of cached content for all files in the selection.

        :param skip_unchanged:
            If ``True`` only inventory of modified files is
            yielded (:py:meth:`flag_modified` must be called beforehand).
        :type skip_unchanged:
            bool

        This generator yields tuples ``((format, path), nuts)`` where ``path``
        is the path to the file, ``format`` is the format assignation or
        ``'detect'`` and ``nuts`` is a list of
        :py:class:`~pyrockoeost.squirrel.model.Nut` objects representing the
        contents of the file.
        '''

        if skip_unchanged:
            where = '''
                WHERE %(db)s.%(file_states)s.file_state == 0
            '''
        else:
            where = ''

        nfiles = execute_get1(self._conn, self._sql('''
            SELECT
                COUNT()
            FROM %(db)s.%(file_states)s
        ''' + where), ())[0]

        def gen():
            sql = self._sql('''
                SELECT
                    %(db)s.%(file_states)s.format,
                    files.path,
                    files.format,
                    files.mtime,
                    files.size,
                    nuts.file_segment,
                    nuts.file_element,
                    kind_codes.kind_id,
                    kind_codes.codes,
                    nuts.tmin_seconds,
                    nuts.tmin_offset,
                    nuts.tmax_seconds,
                    nuts.tmax_offset,
                    kind_codes.deltat
                FROM %(db)s.%(file_states)s
                LEFT OUTER JOIN files
                    ON %(db)s.%(file_states)s.file_id = files.file_id
                LEFT OUTER JOIN nuts
                    ON files.file_id = nuts.file_id
                LEFT OUTER JOIN kind_codes
                    ON nuts.kind_codes_id == kind_codes.kind_codes_id
            ''' + where + '''
                ORDER BY %(db)s.%(file_states)s.file_id
            ''')

            nuts = []
            format_path = None
            db = self.get_database()
            for values in self._conn.execute(sql):
                apath = db.abspath(values[1])
                if format_path is not None and apath != format_path[1]:
                    yield format_path, nuts
                    nuts = []

                format_path = values[0], apath

                if values[2] is not None:
                    nuts.append(model.Nut(
                        values_nocheck=format_path[1:2] + values[2:]))

            if format_path is not None:
                yield format_path, nuts

        return GeneratorWithLen(gen(), nfiles)

    def flag_modified(self, check=True):
        '''
        Mark files which have been modified.

        :param check:
            If ``True`` query modification times of known files on disk. If
            ``False``, only flag unknown files.
        :type check:
            bool

        Assumes file state is 0 for newly added files, 1 for files added again
        to the selection (forces check), or 2 for all others (no checking is
        done for those).

        Sets file state to 0 for unknown or modified files, 2 for known and not
        modified files.
        '''

        db = self.get_database()
        with self.transaction('flag modified') as cursor:
            sql = self._sql('''
                UPDATE %(db)s.%(file_states)s
                SET file_state = 0
                WHERE (
                    SELECT mtime
                    FROM files
                    WHERE
                      files.file_id == %(db)s.%(file_states)s.file_id) IS NULL
                    AND file_state == 1
            ''')

            cursor.execute(sql)

            if not check:

                sql = self._sql('''
                    UPDATE %(db)s.%(file_states)s
                    SET file_state = 2
                    WHERE file_state == 1
                ''')

                cursor.execute(sql)

                return

            def iter_file_states():
                sql = self._sql('''
                    SELECT
                        files.file_id,
                        files.path,
                        files.format,
                        files.mtime,
                        files.size
                    FROM %(db)s.%(file_states)s
                    INNER JOIN files
                        ON %(db)s.%(file_states)s.file_id == files.file_id
                    WHERE %(db)s.%(file_states)s.file_state == 1
                    ORDER BY %(db)s.%(file_states)s.file_id
                ''')

                for (file_id, path, fmt, mtime_db,
                        size_db) in self._conn.execute(sql):

                    path = db.abspath(path)
                    try:
                        mod = io.get_backend(fmt)
                        file_stats = mod.get_stats(path)

                    except FileLoadError:
                        yield 0, file_id
                        continue
                    except io.UnknownFormat:
                        continue

                    if (mtime_db, size_db) != file_stats:
                        yield 0, file_id
                    else:
                        yield 2, file_id

            # could better use callback function here...

            sql = self._sql('''
                UPDATE %(db)s.%(file_states)s
                SET file_state = ?
                WHERE file_id = ?
            ''')

            cursor.executemany(sql, iter_file_states())


__all__ = [
    'Selection',
]



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from .okada import *  # noqa
from .cracksol import *  # noqa

# from pyrockoeost.util import parse_md

# __doc__ = parse_md(__file__)
